/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_dkey.c: DKEY protocol module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_dkey.c,v 1.16 2002/12/15 21:26:17 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include "ssl.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>  
#include <string.h>
#include <fcntl.h>

extern int do_server_estab(aClient *);

#ifndef STATIC_MODULES
Module _module_header = {
	"modules/m_dkey.so",
	"DKEY protocol",
	4,
	"$Revision: 1.16 $",
	0, NULL, NULL
};
void _module_load() {
	addCommand("DKEY", m_dkey, MAXPARA, MSG_UNREGUSE);
	return;
}
void _module_unload() {
	delCommand("DKEY", m_dkey);
	return;
}
#endif

/*
 * m_dkey
 *	parv[0] = sender prefix
 *	parv[1] = command
 *	parv[2] = optional argument
 *	parv[3] = optional argument
 */
int m_dkey(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	if ((cptr->serv == NULL || !(cptr->serv->sflags & SFLAGS_RC4_NEG)) && parc > 1) {
		if (IsPerson(sptr)) {
			return 0;
		}
		return exit_client(sptr, sptr, sptr, "Not negotiating now");
	}
#ifdef HAVE_OPENSSL_ON
	if (!mycmp(parv[1], "START")) {
		char keybuf[1024];

		if (parc != 2) {
			return exit_client(sptr, sptr, sptr, "DKEY START failiure");
		}
		if (sptr->serv->sessioninfo_in != NULL && sptr->serv->sessioninfo_out != NULL) {
			return exit_client(sptr, sptr, sptr, "DKEY START duplicate");
		}

		sptr->serv->sessioninfo_in = dh_start_session();
		sptr->serv->sessioninfo_out = dh_start_session();

		sendto_realops("Initializing Diffie-Hellman key exchange with %s", sptr->name);
	
		dh_get_s_public(keybuf, 1024, sptr->serv->sessioninfo_in);
		sendto_one(sptr, "DKEY PUB I %s", keybuf);

		dh_get_s_public(keybuf, 1024, sptr->serv->sessioninfo_out);
		sendto_one(sptr, "DKEY PUB O %s", keybuf);
		return 0;
	}
	if (!mycmp(parv[1], "PUB")) {
		char keybuf[1024];
		int keylen;

		if (parc != 4 || sptr->serv->sessioninfo_in == NULL || sptr->serv->sessioninfo_out == NULL) {
			return exit_client(sptr, sptr, sptr, "DKEY PUB failiure");
		}
		if (!mycmp(parv[2], "O")) {
			if (!dh_generate_shared(sptr->serv->sessioninfo_in, parv[3])) {
				return exit_client(sptr, sptr, sptr, "DKEY PUB O invalid");
			}
			sptr->serv->dkey_flags |= DKEY_GOTOUT;
		}
		else if (!mycmp(parv[2], "I")) {
			if (!dh_generate_shared(sptr->serv->sessioninfo_out, parv[3])) {
				return exit_client(sptr, sptr, sptr, "DKEY PUB I invalid");
			}
			sptr->serv->dkey_flags |= DKEY_GOTIN;
		}
		else {
			return exit_client(sptr, sptr, sptr, "DKEY PUB bad option");
		}
		if (DKEY_DONE(sptr->serv->dkey_flags)) {
			sendto_one(sptr, "DKEY DONE");
			sptr->serv->sflags |= SFLAGS_RC4_OUT;

			keylen = 1024;
			if (!dh_get_s_shared(keybuf, &keylen, sptr->serv->sessioninfo_in)) {
				return exit_client(sptr, sptr, sptr, "Could not setup encrypted session");
			}
			sptr->serv->rc4_in = rc4_initstate(keybuf, keylen);

			keylen = 1024;
			if (!dh_get_s_shared(keybuf, &keylen, sptr->serv->sessioninfo_out)) {
				return exit_client(sptr, sptr, sptr, "Could not setup encrypted session");
			}
			sptr->serv->rc4_out = rc4_initstate(keybuf, keylen);

			dh_end_session(sptr->serv->sessioninfo_in);
			dh_end_session(sptr->serv->sessioninfo_out);
			sptr->serv->sessioninfo_in = sptr->serv->sessioninfo_out = NULL;
			return 0;
		}
		return 0;
	}
	if (!mycmp(parv[1], "DONE")) {
		if (!((sptr->serv->sessioninfo_in == NULL && sptr->serv->sessioninfo_out == NULL) &&
		  (sptr->serv->rc4_in && sptr->serv->rc4_out))) {
			return exit_client(sptr, sptr, sptr, "DKEY DONE when not done");
		}
		sptr->serv->sflags |= SFLAGS_RC4_IN;
		sendto_realops("Diffie-Hellman key exchange with %s complete, connection encrypted.", sptr->name);
		sendto_one(sptr, "DKEY EXIT");
		return RC4_NEXT_BUFFER;
	}
	if (!mycmp(parv[1], "EXIT")) {
		if (!(cptr->serv->sflags & SFLAGS_RC4_IN) || !(cptr->serv->sflags & SFLAGS_RC4_OUT)) {
			return exit_client(sptr, sptr, sptr, "DKEY EXIT when not in proper stage");
		}
		sptr->serv->sflags &= ~SFLAGS_RC4_NEG;
		return do_server_estab(sptr);
	}
#endif
	return 0;
}
