/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_kline.c: KLINE command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_kline.c,v 1.22 2002/12/15 21:26:18 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include <utmp.h>
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

extern void expire_kline(void *);

#ifndef STATIC_MODULES
Module _module_header = {
	"modules/m_kline.so",
	"/KLINE command",
	4,
	"$Revision: 1.22 $",
	0, NULL, NULL
};
void _module_load() {
	addCommand("KLINE", m_kline, MAXPARA, MSG_NONE);
	return;
}
void _module_unload() {
	delCommand("KLINE", m_kline);
	return;
}
#endif

static void handle_kline_temp(aClient *cptr, char *mask, char *reason)
{
	ConfigItem_ban *ban;

	if (!BadPtr(reason)) {
		if (strlen(reason) > TOPICLEN) {
			reason[TOPICLEN] = '\0';
		}
	}
	else {
		reason = "no reason";
	}
	ban = (ConfigItem_ban *)MyMallocX(sizeof(ConfigItem_ban));
	ban->type = BAN_USERHOST;
	DupString(ban->mask, mask);
	DupString(ban->reason, reason);
	ban->flag.temporary = 1;
	ban->flag.removable = 1;
	ban->flag.remove_on_rehash = 1;
	add_ConfigItem((ConfigItem *)ban, (ConfigItem **)&ci_ban);
	ircdlog(LOG_BAN, "TEMP KLINE on %s (%s) added by %s", ban->mask, ban->reason, get_client_name(cptr, TRUE));
	sendto_realops("%s has added a temporary K:Line on %s (%s).", cptr->name, ban->mask, ban->reason);
	return;
}

static void handle_kline_perm(aClient *cptr, char *mask, time_t expire, char *reason)
{
	ConfigItem_ban *ban;
	int error = 0;

	if (!BadPtr(reason)) {
		if (strlen(reason) > TOPICLEN) {
			reason[TOPICLEN] = '\0';
		}
	}
	else {
		reason = "no reason";
	}
	ban = (ConfigItem_ban *)MyMallocX(sizeof(ConfigItem_ban));
	ban->type = BAN_USERHOST;
	DupString(ban->mask, mask);
	DupString(ban->reason, reason);

	if (WRITE_PERM_KLINES_TO_FILE == 1 && expire == 0) {
		char buffer[1024];
		int fd = -1;

		if ((fd = open(CONFIG_FILE, (O_RDWR|O_APPEND))) == -1) {
			sendto_one(cptr, ":%s NOTICE %s :*** Failed to open %s to write permanent K:Line, "
				"adding with expire-time of %s instead.", me.name, cptr->name, CONFIG_FILE,
				secstodhms(DEFAULT_KLINE_EXPIRE_TIME));
			expire = DEFAULT_KLINE_EXPIRE_TIME;
		}
		else {
			ircsprintf(buffer, "\n/* Permanent K:Line added by %s!%s@%s */\n", cptr->name,
				cptr->username, cptr->host);
			if (write(fd, buffer, strlen(buffer)) > 0) {
				ircsprintf(buffer, "ban userhost {\n\tmask \"%s\";\n\treason \"%s\";\n};\n",
					ban->mask, ban->reason);
				if (write(fd, buffer, strlen(buffer)) <= 0) {
					error = 1;
				}
			}
			else {
				error = 1;
			}
			close(fd);

			if (error == 1) {
				sendto_one(cptr, ":%s NOTICE %s :*** Failed to write permanent K:Line to %s, "
					"adding with expire-time of %s instead.", me.name, cptr->name,
					CONFIG_FILE, secstodhms(DEFAULT_KLINE_EXPIRE_TIME));
				expire = DEFAULT_KLINE_EXPIRE_TIME;
			}
		}
	}
	if ((ban->expire = expire) > 0) {
		char *expiretime = secstodhms(expire);
		ban->setat = timeofday;
		ban->flag.removable = 1;
		addEvent("expire_kline", expire_kline, ban, expire, 0);
		ircdlog(LOG_BAN, "TIMED KLINE on %s (%s) added by %s [expires: %s]", ban->mask, ban->reason,
			get_client_name(cptr, TRUE), expiretime);
		sendto_realops("%s has added a timed K:Line on %s (%s), which expires in %s.",
			cptr->name, ban->mask, ban->reason, expiretime);
	}
	else {
		ban->flag.removable = (WRITE_PERM_KLINES_TO_FILE == 1) ? 0 : 1;
		ban->flag.remove_on_rehash = (WRITE_PERM_KLINES_TO_FILE == 1) ? 1 : 0;
		ircdlog(LOG_BAN, "PERM KLINE on %s (%s) added by %s [%s]", ban->mask, ban->reason,
			get_client_name(cptr, TRUE),
			(WRITE_PERM_KLINES_TO_FILE == 1) ? "written to " CONFIG_FILE : "");
		sendto_realops("%s has added a permanent K:Line on %s (%s)%s.", cptr->name,
			ban->mask, ban->reason,
			(WRITE_PERM_KLINES_TO_FILE == 1) ?
				", which has been written to " CONFIG_FILE : "");
	}
	add_ConfigItem((ConfigItem *)ban, (ConfigItem **)&ci_ban);
	return;
}

/*
 * m_kline
 *	parv[0] = sender prefix
 *	parv[1] = user@host or nickname
 *
 *	Temporary K:Line
 *		parv[2] = reason
 *
 *	Permanent K:Line
 *		parv[2] = expire time (or 0)
 *		parv[3] = reason
 */
int m_kline(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	char *user, *host, *userhost, *p, tmp[HOSTLEN + 2];
	ConfigItem_ban *ban;
	time_t expire = 0;
	aClient *acptr;
	int i = 0;

	if (!OPHasFlag(sptr, OFLAG_KLINE)) {
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		return 0;
	}
	if (parc < 2 || BadPtr(parv[1])) {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "KLINE");
		return 0;
	}
	if ((host = strchr(parv[1], '@')) != NULL || (*parv[1] == '*' || strchr(parv[1], '.') != NULL)) {
		if (host != NULL) {
			user = parv[1];
			*host++ = '\0';
		}
		else {
			user = "*";
			host = parv[1];
		}
		if (strlen(user) > USERLEN) {
			user[USERLEN] = '\0';
		}
		if (strlen(host) > HOSTLEN) {
			host[HOSTLEN] = '\0';
		}
		if (!irccmp(user, "*")) {
			for (p = host; *p != '\0'; p++) {
				if (*p != '*' && *p != '.') {
					i++;
				}
			}
			if (i < 1) {
				sendto_one(sptr, ":%s NOTICE %s :*** What a sweeping K:Line! If only your "
					"admin knew you tried that...", me.name, parv[0]);
				sendto_realops("%s (%s@%s) attempted to K:Line *@*", sptr->name,
					sptr->username, MaskedHost(sptr));
				return 0;
			}
		}
	}
	else {
		if ((acptr = find_chasing(sptr, parv[1], NULL)) == NULL) {
			return 0;
		}
 		if (acptr->user == NULL || IsServer(acptr)) {
			return 0;
		}

		*tmp = '*';
		if (*acptr->username == '~') {
			strcpy(tmp + 1, (char *)acptr->username + 1);
		}
		else {
			strcpy(tmp + 1, acptr->username);
		}
		user = tmp;
		host = acptr->host;
	}

	userhost = make_user_host(user, host);
	for (ban = ci_ban; ban != NULL; ban = (ConfigItem_ban *)ban->next) {
		if (ban->type & BAN_USERHOST && !(ban->type & BAN_USERHOST_S)) {
			if (!match(ban->mask, userhost)) {
				sendto_one(sptr, ":%s NOTICE %s :*** %s is already on the K:Line list as %s (%s)",
					me.name, parv[0], userhost, ban->mask,
					ban->flag.temporary == 1 ? "temporary" : "permanent");
				return 0;
			}
		}
	}
	if (parc >= 4) {
		if ((expire = atime(parv[2])) == 0) {
			if (*parv[2] != '0' || (*parv[2] == '0' && parv[2][1] != '\0')) {
				sendto_one(cptr, ":%s NOTICE %s :*** %s is not a valid expire-time.",
					me.name, cptr->name, parv[2]);
				return 0;
			}
			handle_kline_perm(cptr, userhost, 0, parv[3]);
		}
		else {
			handle_kline_perm(cptr, userhost, expire, parv[3]);
		}
	}
	else if (parc == 3) {
		if ((expire = atime(parv[2])) == 0) {
			if (*parv[2] != '0' || (*parv[2] == '0' && parv[2][1] != '\0')) {
				handle_kline_temp(cptr, userhost, parv[2]);
			}
			else {
				handle_kline_perm(cptr, userhost, 0, NULL);
			}
		}
		else {
			handle_kline_perm(cptr, userhost, expire, NULL);
		}
	}
	else if (parc == 2) {
		handle_kline_temp(cptr, userhost, NULL);
	}
	do_bancheck = 1;
	return 0;
}
