/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_list.c: LIST command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_list.c,v 1.13 2002/12/15 21:26:18 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

static char modebuf[REALMODEBUFLEN], parabuf[REALMODEBUFLEN];

#ifndef STATIC_MODULES
Module _module_header = {
	"modules/m_list.so",
	"/LIST command",
	4,
	"$Revision: 1.13 $",
	0, NULL, NULL
};
void _module_load() {
	addCommand("LIST", m_list, MAXPARA, MSG_NONE);
	return;
}
void _module_unload() {
	delCommand("LIST", m_list);
	return;
}
#endif

/*
 * m_list
 *	parv[0] = sender prefix
 *	parv[1] = channel
 */
int m_list(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	aChannel *chptr;
	time_t currenttime = time(NULL), chantimemin, chantimemax;
	ts_val topictimemin, topictimemax;
	char *name, *p = NULL;
	char modestuff[TOPICLEN + 3 + KEYLEN + 8 + 3 + 1], tmpbuf[3 + KEYLEN + 8 + 3 + 1];
	LOpts *lopt = NULL;
	Link *lp, *yeslist = NULL, *nolist = NULL;
	int usermax, usermin, error = 0, doall = 0;
	static char *usage[] = {
		"   Usage: /RAW LIST <options> (on mIRC) or /QUOTE LIST <options> (on ircII)",
		" ",
		"If you don't include any options, the default is to send you the",
		"entire unfiltered list of channels. Below are the options you can",
		" use, and what channel LIST will return when you use them.",
		" ",
		">number  - List channels with more than <number> people.",
		"<number  - List channels with less than <number> people.",
		"C>number - List channels created between now and <number> minutes ago.",
		"C<number - List channels created earlier than <number> minutes ago.",
		"T>number - List channels whos topics are older than <number> minutes.",
		"T<number - List channels whos topics are not older than <number> minutes.",
		"*mask*   - List channels that match *mask*.",
		"!*mask*  - List channels that do not match *mask*.",
		NULL
	};

	if (cptr != sptr || sptr->user == NULL) {
		return 0;
	}

	if ((lopt = sptr->user->lopt) != NULL) {
		Link *next;
		sendto_one(sptr, rpl_str(RPL_LISTEND), me.name, parv[0]);
		for (lp = lopt->yeslist; lp != NULL; lp = next) {
			next = lp->next;
			MyFree(lp->value.cp);
			free_slink(lp);
		}
		for (lp = lopt->nolist; lp != NULL; lp = next) {
			next = lp->next;
			MyFree(lp->value.cp);
			free_slink(lp);
		}
		MyFree(sptr->user->lopt);
		sptr->user->lopt = NULL;
		return 0;
	}
	if (lifesux > 0) {
		sendto_one(sptr, rpl_str(RPL_LOAD2HI), me.name, sptr->name);
		return 0;
	}
	if (parc < 2 || *parv[1] == '\0') {
		sendto_one(sptr, rpl_str(RPL_LISTSTART), me.name, parv[0]);
		lopt = sptr->user->lopt = (LOpts *)MyMallocX(sizeof(LOpts));
		lopt->showall = 1;
		if (DBufLength(&cptr->sendQ) < 2048) {
			send_list(cptr, 64);
		}
		return 0;
	}
	if (parc == 2 && parv[1][0] == '?' && parv[1][1] == '\0') {
		char **ptr = usage;
		for (; *ptr; ptr++) {
			sendto_one(sptr, rpl_str(RPL_COMMANDSYNTAX), me.name, parv[0], *ptr);
		}
		return 0;
	}

	sendto_one(sptr, rpl_str(RPL_LISTSTART), me.name, parv[0]);

	chantimemax = topictimemax = currenttime + 86400;
	chantimemin = topictimemin = 0;
	usermin = 1;
	usermax = -1;
	for (name = strtoken(&p, parv[1], ","); name != NULL && error == 0; name = strtoken(&p, NULL, ",")) {
		switch (*name) {
			case '<':
				usermax = atoi(name + 1) - 1;
				doall = 1;
				break;
			case '>':
				usermin = atoi(name + 1) + 1;
				doall = 1;
				break;
			case 'C':
			case 'c':
				name++;
				switch (*name++) {
					case '<':
						chantimemax = currenttime - 60 * atoi(name);
						doall = 1;
						break;
					case '>':
						chantimemin = currenttime - 60 * atoi(name);
						doall = 1;
						break;
					default:
						sendto_one(sptr, err_str(ERR_LISTSYNTAX), me.name,
							parv[0]);
						error = 1;
				}
				break;
			case 'T':
			case 't':
				name++;
				switch (*name++) {
					case '<':
						topictimemax = currenttime - 60 * atoi(name);
						doall = 1;
						break;
					case '>':
						topictimemin = currenttime - 60 * atoi(name);
						doall = 1;
						break;
					default:
						sendto_one(sptr, err_str(ERR_LISTSYNTAX), me.name,
							parv[0]);
						error = 1;
				}
				break;
			default:
				if (*name == '!') {
					doall = 1;
					lp = make_slink();
					lp->next = nolist;
					nolist = lp;
					DupString(lp->value.cp, name + 1);
				}
				else if (strchr(name, '*') != NULL || strchr(name, '*') != NULL) {
					doall = 1;
					lp = make_slink();
					lp->next = yeslist;
					nolist = lp;
					DupString(lp->value.cp, name);
				}
				else {
					if ((chptr = find_channel(name, NULL)) != NULL) {
						if (HasMode(sptr, UMODE_SADMIN)) {
							*modebuf = *parabuf = '\0';
							channel_modes(sptr, modebuf, parabuf, chptr);

							ircsprintf(tmpbuf, " [%s%s%s]", modebuf,
								*parabuf ? " " : "",
								*parabuf ? parabuf : "");
							ircsprintf(modestuff, "%-2s %s", tmpbuf, chptr->topic);

							sendto_one(sptr, rpl_str(RPL_LIST), me.name, parv[0],
								name, chptr->users, modestuff);
						}
						else if (ShowChannel(sptr, chptr)) {
							sendto_one(sptr, rpl_str(RPL_LIST), me.name,
								parv[0], name, chptr->users,
								chptr->topic);
						}
					}
				}
		}
	}
	if (doall == 1) {
		lopt = sptr->user->lopt = (LOpts *)MyMallocX(sizeof(LOpts));
		lopt->usermax = usermax;
		lopt->usermin = usermin;
		lopt->chantimemax = chantimemax;
		lopt->chantimemin = chantimemin;
		lopt->topictimemax = topictimemax;
		lopt->topictimemin = topictimemin;
		lopt->yeslist = yeslist;
		lopt->nolist = nolist;

		if (DBufLength(&cptr->sendQ) < 2048) {
			send_list(cptr, 64);
		}
		return 0;
	}
	sendto_one(sptr, rpl_str(RPL_LISTEND), me.name, parv[0]);
	return 0;
}
