/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_oper.c: OPER command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_oper.c,v 1.35 2002/12/15 21:26:18 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

static ErrorMsg errors[] = {
	{ ERR_NOOPERHOST, "NO CONFIGURATION" },
	{ ERR_NOOPERHOST, "USERHOST MISMATCH" },
	{ ERR_PASSWDMISMATCH, "AUTHENTICATION FAILURE" }
};

#ifndef STATIC_MODULES
Module _module_header = {
	"modules/m_oper.so",
	"/OPER command",
	4,
	"$Revision: 1.35 $",
	0, NULL, NULL
};
void _module_load() {
	addCommand("OPER", m_oper, MAXPARA, MSG_NONE);
	return;
}
void _module_unload() {
	delCommand("OPER", m_oper);
	return;
}
#endif

/*
 * m_oper
 *	parv[0] = sender prefix
 *	parv[1] = login
 *	parv[2] = password
 */
int m_oper(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	ConfigItem_oper *oper = NULL;
	int old = 0, send_error = 0;
	ErrorMsg *err = NULL;
	char *from;

	if (!IsServer(cptr) && (BadPtr(parv[1]) || BadPtr(parv[2]))) {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "OPER");
		return 0;
	}
	if ((IsServer(cptr) || IsMe(cptr)) && !HasMode(sptr, UMODE_OPER)) {
		AddMode(sptr, UMODE_OPER);
		sendto_serv_butone(cptr, ":%s MODE %s :+o", parv[0], parv[0]);
		Count.oper++;
		if (IsMe(cptr)) {
			sendto_one(sptr, rpl_str(RPL_YOUREOPER), me.name, parv[0]);
		}
		return 0;
	}
	else if (HasMode(sptr, UMODE_OPER)) {
		if (MyClient(sptr)) {
			sendto_one(sptr, rpl_str(RPL_YOUREOPER), me.name, parv[0]);
		}
		return 0;
	}

	from = make_user_host(sptr->username, sptr->host);
	for (oper = ci_oper; oper != NULL; oper = (ConfigItem_oper *)oper->next) {
		if (irccmp(oper->name, parv[1])) {
			continue;
		}
		if (!source_check((Source **)&oper->source, from)) {
			send_error = 1;
			continue;
		}
		if (!auth_check(oper->auth, parv[2])) {
			send_error = 2;
			continue;
		}
		break;
	}
	if (oper == NULL) {
		err = &errors[send_error];
		ircdlog(LOG_OPER, "failed oper attempt by %s [%s]", get_client_name(cptr, FALSE),
			err->message);
		sendto_one(sptr, err_str(err->numeric), me.name, parv[0]);
		if (FAILOPERNOTICE == 1) {
			sendto_realops_lev(SPY_LEV, "Failed OPER attempt by %s (%s@%s) [%s]", sptr->name,
				sptr->username, MaskedHost(sptr), err->message);
		}
		return 0;
	}
	ircdlog(LOG_OPER, "%s successfully opered as %s", get_client_name(sptr, FALSE), oper->name);

	if (sptr->class != NULL && sptr->class->clients > 0) {
		sptr->class->clients--;
	}
	sptr->class = oper->class;
	sptr->class->clients++;
	sptr->pingval = sptr->class->pingtime;
	old = (sptr->umode & ALL_UMODES);

	OPAddFlag(sptr, oper->options);
	if (OPHasFlag(sptr, OFLAG_SADMIN)) {
		AddMode(sptr, UMODE_SADMIN);
	}
	if (OPHasFlag(sptr, OFLAG_RSTAFF)) {
		AddMode(sptr, UMODE_RSTAFF);
	}
	if (OPHasFlag(sptr, OFLAG_NETADMIN)) {
		AddMode(sptr, UMODE_NETADMIN);
	}
	if (OPHasFlag(sptr, OFLAG_ADMIN)) {
		AddMode(sptr, UMODE_ADMIN);
	}
	if (OPHasFlag(sptr, OFLAG_COADMIN)) {
		AddMode(sptr, UMODE_COADMIN);
	}
	if (OPHasFlag(sptr, OFLAG_FNOTICE)) {
		AddMode(sptr, UMODE_FNOTICE);
	}
	if (OPHasFlag(sptr, OFLAG_LCLICONN)) {
		AddMode(sptr, UMODE_LCLICONN);
	}
	if (OPHasFlag(sptr, OFLAG_GCLICONN)) {
		AddMode(sptr, UMODE_GCLICONN);
	}
	AddMode(sptr, OPER_UMODES);
	Count.oper++;

	send_umode_out(cptr, sptr, old);
	addto_fdlist(sptr->fd, &oper_fdlist);
	sendto_one(sptr, rpl_str(RPL_YOUREOPER), me.name, parv[0]);

	sendto_ops("%s (%s@%s) is now %s [%s]", sptr->name, sptr->username, MaskedHost(sptr),
		OPHasFlag(sptr, OFLAG_NETADMIN) ? "a Network Administrator" :
		OPHasFlag(sptr, OFLAG_ADMIN) ? "a Server Administrator" :
		OPHasFlag(sptr, OFLAG_COADMIN) ? "a Server Co-Administrator" : "an Operator",
		oper->name);
	if (HasMode(sptr, UMODE_NETADMIN)) {
		sendto_serv_butone(&me, ":%s GLOBOPS :%s (%s@%s) is now a Network Administrator",
			me.name, sptr->name, sptr->username, MaskedHost(sptr));
	}
	if (OPHasFlag(sptr, OFLAG_OPERMOTD)) {
		send_message_file(sptr, &opermotd);
	}

	mask_client(sptr, maskme(sptr));
	if (!BadPtr(oper->autojoinchans)) {
		char *chans[3] = { sptr->name, oper->autojoinchans, NULL };
		runHook4(HOOKTYPE_JOIN, sptr, sptr, 3, chans);
	}
	return 0;
}
