/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_sjoin.c: SJOIN protocol module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_sjoin.c,v 1.36 2002/12/15 21:26:19 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

typedef struct parv {
	int parc;
	char *parv[256];
} aParv;

extern void set_mode(aChannel *, aClient *, int, char **, u_int *, char [MAXMODEPARAMS][MODEBUFLEN + 3], int);
extern chanMember *add_user_to_channel(aChannel *, aClient *, int);
extern aChannel *get_channel(aClient *, char *, int);
extern aCtab cFlagTab[];

static char modebuf[REALMODEBUFLEN];
static char parabuf[REALMODEBUFLEN];
static char parabuf2[REALMODEBUFLEN];
static char buf[BUFSIZE];
static char sjbuf[BUFSIZE];
static aParv pparv;
static int mbufpos = 0;
static int pbufpos = 0;
static int pbufcnt = 0;
static int sjbufpos = 0;

#ifndef STATIC_MODULES
Module _module_header = {
	"modules/m_sjoin.so",
	"SJOIN protocol",
	4,
	"$Revision: 1.36 $",
	0, NULL, NULL
};
void _module_load() {
	addCommand("SJOIN", m_sjoin, MAXPARA, MSG_NONE);
	return;
}
void _module_unload(void) {
	delCommand("SJOIN", m_sjoin);
	return;
}
#endif

static inline aParv *mp2parv(char *mbuf, char *pbuf)
{
	char *p, *s;
	int c = 1;

	pparv.parv[0] = mbuf;
	for (s = strtoken(&p, pbuf, " "); s != NULL; s = strtoken(&p, NULL, " ")) {
		pparv.parv[c++] = s;
	}
	pparv.parv[c] = '\0';
	return &pparv;
}

static inline void sj_sendit(aClient *cptr, aChannel *chptr, char *from)
{
	if (mbufpos > 1) {
		modebuf[mbufpos] = '\0';
		parabuf[pbufpos] = '\0';
		sendto_channel_butserv(chptr, cptr, ":%s MODE %s %s %s", from, chptr->chname, modebuf, parabuf);
	}
	modebuf[(mbufpos = 0)] = '\0';
	parabuf[(pbufpos = 0)] = '\0';
	return;
}

static inline void sj_senditNT(aClient *cptr, aChannel *chptr, char *from)
{
	sendto_channel_butserv(chptr, cptr, ":%s MODE %s %s %s", from, chptr->chname, modebuf, parabuf);
	*modebuf = '\0';
	*parabuf = '\0';
	return;
}

static inline int sj_senditmax(aClient *cptr, aChannel *chptr, char *from)
{
	if (pbufcnt >= MAXMODEPARAMS) {
		modebuf[mbufpos] = '\0';
		parabuf[pbufpos] = '\0';
		sendto_channel_butserv(chptr, cptr, ":%s MODE %s %s %s", from, chptr->chname, modebuf, parabuf);
		modebuf[(mbufpos = 0)] = '\0';
		parabuf[(pbufpos = 0)] = '\0';
		pbufcnt = 0;
		return 1;
	}
	return 0;
}

static inline void sj_addmode(char modechar)
{
	modebuf[mbufpos++] = modechar;
	return;
}

static inline void sj_addpara(char *para)
{
	char *ptr;
	if (pbufpos > 0) {
		parabuf[pbufpos++] = ' ';
	}
	for (ptr = para; *ptr; ptr++) {
		parabuf[pbufpos++] = *ptr;
	}
	parabuf[pbufpos] = '\0';
	pbufcnt++;
	return;
}

static inline void sj_addtobuf(char *string)
{
	char *ptr;
	int pos = 0;

	*buf = '\0';
	for (ptr = string; *ptr != '\0'; ptr++) {
		buf[pos++] = *ptr;
	}
	buf[pos] = '\0';
	return;
}

static inline void sj_addtosjbuf(char *nick)
{
	char *ptr;
	if (sjbufpos > 0) {
		sjbuf[sjbufpos++] = ' ';
	}
	for (ptr = nick; *ptr != '\0'; ptr++) {
		sjbuf[sjbufpos++] = *ptr;
	}
	sjbuf[sjbufpos] = '\0';
	return;
}

/*
 * m_sjoin
 *	parv[0] = sender prefix
 *	parv[1] = timestamp
 *	parv[2] = channel name
 *	if (parc == 3) - Client SJOIN
 *	if (parc > 3)  - Server SJOIN
 *		parv[3] = modes
 *		parv[4 to parc - 2] = mode parameters
 *		parv[parc - 1] = user list with statuses and channel bans
 */
int m_sjoin(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	int merge = 0, pcount, fl = 0, ckeylen = 0, okeylen = 0, break_loop = 0, i, people = 0;
	int nomodes = 0, noparams = 0, wipeours = 0, wipetheirs = 0, isnew = 0;
	char pvar[MAXMODEPARAMS][MODEBUFLEN + 3], *s, *p, *ns;
	aChannel *chptr;
	chanMember *cm = NULL;
	aClient *acptr;
	Mode oldmode;
	aParv *ap;
	aCtab *tab;
	aBan *ban;
	ts_val ts;

	if (MyClient(sptr) || (parc < 6 && IsServer(sptr)) || (parc < 3 && IsPerson(sptr))) {
		return 0;
	}
	if (!IsChannelName(parv[2])) {
		return 0;
	}
	if (parc > 3) {
		if (parv[3][1] == '\0') {
			nomodes = 1;
		}
		if (!ircncmp(parv[4], "<none>", 6)) {
			noparams = 1;
		}
	}

	isnew = ChannelExists(parv[2]) ? 0 : 1;
	chptr = get_channel(cptr, parv[2], CREATE);
	ts = atol(parv[1]);

	if (parc == 3) {
		if (isnew == 1) {
			chptr->channelts = ts;
		}
		else if (ts == 0 || chptr->channelts == 0) {
			chptr->channelts = 0;
		}
		else if (ts < chptr->channelts) {
			chptr->channelts = ts;
		}
		if (!IsMember(sptr, chptr)) {
			add_user_to_channel(chptr, sptr, 0);
			sendto_channel_butserv(chptr, sptr, ":%s JOIN :%s", sptr->name, chptr->chname);
		}
		sendto_ssjoin_servs_butone(0, chptr, cptr, ":%s SJOIN %ld %s", sptr->name, chptr->channelts,
			chptr->chname);
		sendto_ssjoin_servs_butone(1, chptr, cptr, ":%s SJOIN %ld %s", sptr->name, chptr->channelts,
			chptr->chname);
		return 0;
	}
	if (chptr->channelts > ts) {
		wipeours = 1;
		sendto_channel_butserv(chptr, &me, ":%s NOTICE %s :*** Notice -- TS for %s changed "
			"from %lu to %lu", me.name, chptr->chname, chptr->chname,
			chptr->channelts, ts);
		chptr->channelts = ts;
	}
	else if (chptr->channelts < ts && chptr->channelts != 0) {
		wipetheirs = 1;
	}
	else if (chptr->channelts == ts) {
		merge = 1;
	}
	if (chptr->channelts == 0) {
		chptr->channelts = ts;
	}

	modebuf[(mbufpos = 0)] = '\0';
	parabuf[(pbufpos = 0)] = '\0';
	pbufcnt = 0;
	channel_modes(cptr, modebuf, parabuf, chptr);

	if (wipeours == 1) {
		if (modebuf[1] != '\0') {
			*modebuf = '-';
			ap = mp2parv(modebuf, parabuf);
			set_mode(chptr, cptr, ap->parc, ap->parv, &pcount, pvar, 0);
			sj_senditNT(cptr, chptr, me.name);
		}

		sj_addmode('-');
		for (cm = chptr->members; cm != NULL; cm = cm->nextuser) {
			if (cm->flags & CHFL_CHANADMIN) {
				cm->flags &= ~CHFL_CHANADMIN;
				sj_addmode('a');
				sj_addpara(cm->cptr->name);
			}
			if (cm->flags & CHFL_CHANOP) {
				cm->flags &= ~CHFL_CHANOP;
				sj_addmode('o');
				sj_addpara(cm->cptr->name);
			}
			if (cm->flags & CHFL_HALFOP) {
				cm->flags &= ~CHFL_HALFOP;
				sj_addmode('o');
				sj_addpara(cm->cptr->name);
			}
			if (cm->flags & CHFL_VOICE) {
				cm->flags &= ~CHFL_VOICE;
				sj_addmode('v');
				sj_addpara(cm->cptr->name);
			}
			if (sj_senditmax(cptr, chptr, me.name)) {
				sj_addmode('-');
			}
		}
		for (ban = chptr->banlist; ban; ban = ban->next) {
			sj_addtobuf(ban->banstr);
			if (del_ban_fromlist(chptr, ban->banstr, (aBan **)&chptr->banlist, 1)) {
				sj_addmode('b');
				sj_addpara(buf);
			}
			if (sj_senditmax(cptr, chptr, me.name)) {
				sj_addmode('-');
			}
		}
		for (ban = chptr->exbanlist; ban; ban = ban->next) {
			sj_addtobuf(ban->banstr);
			if (del_ban_fromlist(chptr, ban->banstr, (aBan **)&chptr->exbanlist, 0)) {
				sj_addmode('b');
				sj_addpara(buf);
			}
			if (sj_senditmax(cptr, chptr, me.name)) {
				sj_addmode('-');
			}
		}
		for (ban = chptr->invexlist; ban; ban = ban->next) {
			sj_addtobuf(ban->banstr);
			if (del_ban_fromlist(chptr, ban->banstr, (aBan **)&chptr->invexlist, 0)) {
				sj_addmode('I');
				sj_addpara(buf);
			}
			if (sj_senditmax(cptr, chptr, me.name)) {
				sj_addmode('-');
			}
		}
		sj_sendit(cptr, chptr, me.name);

		if (*chptr->topic != '\0') {
			chptr->topic[0] = '\0';
			sendto_channel_butserv(chptr, cptr, ":%s TOPIC %s :%s", me.name,
				chptr->chname, chptr->topic);
		}
	}

	sjbuf[(sjbufpos = 0)] = '\0';
	sj_addmode('+');
	sj_addtobuf(parv[parc - 1]);
	for (s = strtoken(&p, buf, " "); s; s = strtoken(&p, NULL, " ")) {
		break_loop = 0;
		fl = 0;
		for (ns = s; (*ns == '*' || *ns == '@' || *ns == '%' || *ns == '+' ||
		  *ns == '&' || *ns == '"') && break_loop == 0; ns++) {
			switch (*ns) {
				case '*':
					fl |= CHFL_CHANADMIN;
					break;
				case '@':
					fl |= CHFL_CHANOP;
					break;
				case '%':
					fl |= CHFL_HALFOP;
					break;
				case '+':
					fl |= CHFL_VOICE;
					break;
				case '&':
					fl |= MODE_BAN;
					break_loop = 1;
					break;
				case '"':
					fl |= MODE_EXBAN;
					break_loop = 1;
					break;
				case '$':
					fl |= MODE_INVEX;
					break_loop = 1;
					break;
			}
		}
		if (fl & MODE_BAN) {
			if (wipetheirs == 0) {
				if (add_ban_tolist(sptr, chptr, ns, (aBan **)&chptr->banlist, 1)) {
					people++;
					sj_addmode('b');
					sj_addpara(ns);
					sj_addtosjbuf(s);
				}
			}
		}
		else if (fl & MODE_EXBAN) {
			if (wipetheirs == 0) {
				if (add_ban_tolist(sptr, chptr, ns, (aBan **)&chptr->exbanlist, 0)) {
					people++;
					sj_addmode('e');
					sj_addpara(ns);
					sj_addtosjbuf(s);
				}
			}
		}
		else if (fl & MODE_INVEX) {
			if (wipetheirs == 0) {
				if (add_ban_tolist(sptr, chptr, ns, (aBan **)&chptr->invexlist, 0)) {
					people++;
					sj_addmode('I');
					sj_addpara(ns);
					sj_addtosjbuf(s);
				}
			}
		}
		else {
			if (wipetheirs == 1) {
				fl = 0;
			}
			if ((acptr = find_chasing(sptr, ns, NULL)) == NULL) {
				sendto_realops_lev(SKILL_LEV, "Removing missing user %s in SJOIN %s to avoid desynch",
					ns, chptr->chname);
				sendto_one(cptr, ":%s KILL %s :Missing user in SJOIN for %s", me.name, ns,
					chptr->chname);
				continue;
			}
			if (acptr->from != cptr) {
				sendto_one(cptr, ":%s KICK %s %s :(SJOIN) fake direction", me.name, chptr->chname,
					acptr->name);
				continue;
			}
			if (!IsMember(acptr, chptr)) {
				cm = add_user_to_channel(chptr, acptr, fl);
				sendto_channel_butserv(chptr, acptr, ":%s JOIN :%s", acptr->name,
					chptr->chname);
			}
			if (cm == NULL) {
				if ((cm = find_user_member(acptr->user->channel, chptr)) == NULL) {
					sendto_realops("DESYNCH! Missing chanMember link for user %s in SJOIN for "
						"channel %s", acptr->name, chptr->chname);
					continue;
				}
			}
			if (fl & CHFL_CHANADMIN) {
				cm->flags |= CHFL_CHANADMIN;
				sj_addmode('a');
				sj_addpara(acptr->name);
			}
			if (fl & CHFL_CHANOP) {
				cm->flags |= CHFL_CHANOP;
				sj_addmode('o');
				sj_addpara(acptr->name);
			}
			if (fl & CHFL_HALFOP) {
				cm->flags |= CHFL_HALFOP;
				sj_addmode('h');
				sj_addpara(acptr->name);
			}
			if (fl & CHFL_VOICE) {
				cm->flags |= CHFL_VOICE;
				sj_addmode('v');
				sj_addpara(acptr->name);
			}
			people++;
			sj_addtosjbuf(s);
		}
		if (sj_senditmax(cptr, chptr, me.name)) {
			sj_addmode('+');
		}
	}
	sj_sendit(cptr, chptr, me.name);

	modebuf[(mbufpos = 0)] = '\0';
	parabuf[(pbufpos = 0)] = '\0';
	pbufcnt = 0;

	if (merge == 0 && wipetheirs == 0 && nomodes == 0) {
		strcpy(modebuf, parv[3]);
		if (noparams == 0) {
			for (i = 4; i <= parc - 2; i++) {
				sj_addpara(parv[i]);
			}
		}
		strcpy(parabuf2, parabuf);
		ap = mp2parv(modebuf, parabuf2);
		set_mode(chptr, cptr, ap->parc, ap->parv, &pcount, pvar, 0);
		sj_senditNT(cptr, chptr, me.name);
	}
	if (merge == 1 && nomodes == 0) {
		strcpy(modebuf, parv[3]);
		if (noparams == 1) {
			ap = mp2parv(modebuf, parabuf);
			set_mode(chptr, cptr, ap->parc, ap->parv, &pcount, pvar, 0);
			sj_senditNT(cptr, chptr, me.name);
		}
		else {
			memcpy(&oldmode, &chptr->mode, sizeof(Mode));
			for (i = 4; i <= parc - 2; i++) {
				sj_addpara(parv[i]);
			}
			ap = mp2parv(modebuf, parabuf);
			set_mode(chptr, cptr, ap->parc, ap->parv, &pcount, pvar, 0);

			modebuf[(mbufpos = 0)] = '\0';
			parabuf[(pbufpos = 0)] = '\0';
			pbufcnt = 0;
			sj_addmode('-');

			if (oldmode.limit != 0 && chptr->mode.limit == 0) {
				sj_addmode('l');
				sj_addpara((char *)my_itoa(oldmode.limit));
			}
			if (*oldmode.key != '\0' && *chptr->mode.key == '\0') {
				sj_addmode('k');
				sj_addpara(oldmode.key);
			}
			for (tab = &cFlagTab[0]; tab->mode != 0; tab++) {
				if ((oldmode.mode & tab->mode) && !(chptr->mode.mode & tab->mode) && tab->param == 0) {
					sj_addmode(tab->flag);
				}
			}
			if (mbufpos > 1) {
				sj_addmode('+');
			}
			else {
				*modebuf = '+';
			}
			if (oldmode.limit == 0 && chptr->mode.limit != 0) {
				sj_addmode('l');
				sj_addpara((char *)my_itoa(chptr->mode.limit));
			}
			if (*oldmode.key == '\0' && *chptr->mode.key != '\0') {
				sj_addmode('k');
				sj_addpara(chptr->mode.key);
			}
			if (oldmode.limit != 0 && chptr->mode.limit != 0 && oldmode.limit != chptr->mode.limit) {
				chptr->mode.limit = oldmode.limit > chptr->mode.limit ? oldmode.limit : chptr->mode.limit;
				if (oldmode.limit != chptr->mode.limit) {
					sj_addmode('l');
					sj_addpara((char *)my_itoa(chptr->mode.limit));
				}
			}
			if (*oldmode.key != '\0' && *chptr->mode.key != '\0' && irccmp(oldmode.key, chptr->mode.key)) {
				okeylen = strlen(oldmode.key);
				ckeylen = strlen(chptr->mode.key);
				if (okeylen == ckeylen) {
					sendto_channel_butserv(chptr, sptr, ":%s NOTICE %s :WARNING! "
						"Unable to determine which key (+k) to use in SJOIN "
						"for %s, removing all keys.", me.name, chptr->chname,
						chptr->chname);
					okeylen = ckeylen = -1;
				}
				else if (okeylen > ckeylen) {
					strcpy(chptr->mode.key, oldmode.key);
				}
				else {
					sj_addmode('k');
					sj_addpara(chptr->mode.key);
				}
			}
			sj_sendit(cptr, chptr, me.name);

			if (okeylen == -1 && ckeylen == -1) {
				sj_addmode('-');
				sj_addmode('k');
				sj_addpara(chptr->mode.key);
				ap = mp2parv(modebuf, parabuf);
				set_mode(chptr, cptr, ap->parc, ap->parv, &pcount, pvar, 0);
				sj_sendit(cptr, chptr, me.name);
			}
		}
	}
	if (noparams == 0) {
		for (i = 4; i <= parc - 2; i++) {
			sj_addpara(parv[i]);
		}
	}
	if (pbufcnt > 0) {
		sendto_ssjoin_servs_butone(0, chptr, cptr, ":%s SJOIN %ld %s %s %s :%s", sptr->name, ts,
			chptr->chname, parv[3], parabuf, sjbuf);
		sendto_ssjoin_servs_butone(1, chptr, cptr, ":%s SJOIN %ld %s %s %s :%s", sptr->name, ts,
			chptr->chname, parv[3], parabuf, sjbuf);
	}
	else {
		sendto_ssjoin_servs_butone(0, chptr, cptr, ":%s SJOIN %ld %s %s :%s", sptr->name, ts,
			chptr->chname, parv[3], sjbuf);
		sendto_ssjoin_servs_butone(1, chptr, cptr, ":%s SJOIN %ld %s %s <none> :%s", sptr->name,
			ts, chptr->chname, parv[3], sjbuf);
	}
	return 0;
}
