/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_who.c: WHO command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_who.c,v 1.18 2002/12/15 21:26:20 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

static int (*nchkfn)(char *, char *);
static int (*uchkfn)(char *, char *);
static int (*hchkfn)(char *, char *);
static int (*gchkfn)(char *, char *);
static int (*ichkfn)(char *, char *);
static SOpts wsopts;

extern int user_modes[];

#ifndef STATIC_MODULES
Module _module_header = {
	"modules/m_who.so",
	"/WHO command",
	4,
	"$Revision: 1.18 $",
	0, NULL, NULL
};
void _module_load() {
	addCommand("WHO", m_who, MAXPARA, MSG_NONE);
	return;
}
void _module_unload(void) {
	delCommand("WHO", m_who);
	return;
}
#endif

static int build_searchopts(aClient *sptr, int parc, char *parv[])
{
	static char *who_help[] = {
		"/WHO [+|-][acghimnsuCM] [args]",
		"Flags are specified like channel modes, the flags cgmnsu all have arguments.",
		"Flags are set to a positive check by +, and a negative check by -. The flags",
		"are as follows:",
		" ",
		"a - user is away",
		"c <channel> - user is on <channel>, no wildcards accepted",
		"g <realname> - user has string <realname> in their GCOS (realname), wildcards",
		"               accepted, opers only",
		"h <host> - user has string <host> in their hostname/maskedhost, wildcards accepted",
		"i <ip> - user is from <ip>, wildcards accepted, opers only",
		"m <usermodes> - user has <usermodes> set on them, only o/a for nonopers",
		"s <server> - user is on server <server>, wildcards not accepted",
		"u <user> - user has string <user> in their username, wildcards accepted",
		" ",
		"Behaviour flags:",
		" ",
		"C - show first visible channel user is in",
		"M - check for user in channels I am a member of",
		NULL
	};
	char *flags, change = 1, *s;
	int args = 1, i;

	memset((char *)&wsopts, '\0', sizeof(SOpts));
	if (parc < 1 || *parv[0] == '?') {
		char **ptr;
		for (ptr = who_help; *ptr != NULL; ptr++) {
			sendto_one(sptr, rpl_str(RPL_COMMANDSYNTAX), me.name, sptr->name, *ptr);
		}
		sendto_one(sptr, rpl_str(RPL_ENDOFWHO), me.name, sptr->name, "?");
		return 0;
	}
	else if (*parv[0] == '0' && parv[0][1] == 0) {
		if (parc > 1 && *parv[1] == 'o') {
			wsopts.check_umode = 1;
			wsopts.umode_plus = 1;
			wsopts.umodes = UMODE_OPER;
		}
		wsopts.host_plus = 1;
		wsopts.host = "*";
		return 1;
	}
	else if (*parv[0] != '+' && *parv[0] != '-') {
		if (*parv[0] == '#' || *parv[0] == '&') {
			if ((wsopts.channel = find_channel(parv[0], NULL)) == NULL) {
				sendto_one(sptr, err_str(ERR_NOSUCHCHANNEL), me.name, sptr->name,
					parv[0]);
				return 0;
			}
		}
		else {
			if (strchr(parv[0], '.') != NULL) {
				wsopts.host_plus = 1;
				wsopts.host = parv[0];
			}
			else {
				wsopts.nick_plus = 1;
				wsopts.nick = parv[0];
			}
		}
		return 1;
	}

	flags = parv[0];
	while (*flags) {
		switch (*flags) {
			case '+':
			case '-':
				change = *flags == '+' ? 1 : 0;
				break;
			case 'a':
				wsopts.away_plus = (change == 1) ? 1 : 0;
				wsopts.check_away = 1;
				break;
			case 'C':
				wsopts.show_chan = change;
				break;
			case 'M':
				wsopts.search_chan = change;
				break;
			case 'c':
				if (parv[args] == NULL || change == 0) {
					sendto_one(sptr, err_str(ERR_WHOSYNTAX), me.name, sptr->name);
					return 0;
				}
				if ((wsopts.channel = find_channel(parv[args], NULL)) == NULL) {
					sendto_one(sptr, err_str(ERR_NOSUCHCHANNEL), me.name, sptr->name,
						parv[args]);
					return 0;
				}
				wsopts.chan_plus = change;
				args++;
				break;
			case 'g':
				if (parv[args] == NULL || !HasMode(sptr, UMODE_OPER)) {
					sendto_one(sptr, err_str(ERR_WHOSYNTAX), me.name, sptr->name);
					return 0;
				}
				wsopts.gcos = parv[args];
				wsopts.gcos_plus = change;
				args++;
				break;
			case 'h':
				if (parv[args] == NULL) {
					sendto_one(sptr, err_str(ERR_WHOSYNTAX), me.name, sptr->name);
					return 0;
				}
				wsopts.host = parv[args];
				wsopts.host_plus = change;
				args++;
				break;
			case 'i':
				if (parv[args] == NULL || !HasMode(sptr, UMODE_OPER)) {
					sendto_one(sptr, err_str(ERR_WHOSYNTAX), me.name, sptr->name);
					return 0;
				}
				wsopts.ip = parv[args];
				wsopts.ip_plus = change;
				args++;
				break;
			case 'm':
				if (parv[args] == NULL) {
					sendto_one(sptr, err_str(ERR_WHOSYNTAX), me.name, sptr->name);
					return 0;
				}
				for (s = parv[args]; *s != '\0'; s++) {
					for (i = 1; user_modes[i] != 0x0; i += 2) {
						if (*s == (char)user_modes[i]) {
							wsopts.umodes |= user_modes[i - 1];
							break;
						}
					}
				}
				if (!HasMode(sptr, UMODE_OPER)) {
					wsopts.umodes = (wsopts.umodes & (UMODE_OPER|UMODE_SADMIN));
				}
				wsopts.umode_plus = change;
				if (wsopts.umodes == 1) {
					wsopts.check_umode = 1;
				}
				args++;
				break;
			case 'n':
				if (parv[args] == NULL) {
					sendto_one(sptr, err_str(ERR_WHOSYNTAX), me.name, sptr->name);
					return 0;
				}
				wsopts.nick = parv[args];
				wsopts.nick_plus = change;
				args++;
				break;
			case 's':
				if (parv[args] == NULL || change == 0) {
					sendto_one(sptr, err_str(ERR_WHOSYNTAX), me.name, sptr->name);
					return 0;
				}
				if ((wsopts.server = find_server(parv[args], NULL)) == NULL) {
					sendto_one(sptr, err_str(ERR_NOSUCHSERVER), me.name, sptr->name,
						parv[args]);
					return 0;
				}
				wsopts.serv_plus = change;
				args++;
				break;
			case 'u':
				if (parv[args] == NULL) {
					sendto_one(sptr, err_str(ERR_WHOSYNTAX), me.name, sptr->name);
					return 0;
				}
				wsopts.user = parv[args];
				wsopts.user_plus = change;
				args++;
				break;
			default:
				args++;
				break;
		}
		flags++;
	}

	if (wsopts.search_chan == 1 && !(wsopts.check_away == 1 || wsopts.gcos != NULL ||
	  wsopts.host != NULL || wsopts.check_umode == 1 || wsopts.server != NULL ||
	  wsopts.user != NULL)) {
		if (parv[args] == NULL || wsopts.channel != NULL || wsopts.nick != NULL ||
		  *parv[args] == '#' || *parv[args] == '&') {
			sendto_one(sptr, err_str(ERR_WHOSYNTAX), me.name, sptr->name);
			return 0;
		}
		if (strchr(parv[args], '.') != NULL) {
			wsopts.host_plus = 1;
			wsopts.host = parv[args];
		}
		else {
			sendto_one(sptr, err_str(ERR_WHOSYNTAX), me.name, sptr->name);
			return 0;
		}
	}
	else if (wsopts.show_chan == 1 && !(wsopts.check_away == 1 || wsopts.gcos != NULL ||
	  wsopts.host != NULL || wsopts.check_umode == 1 || wsopts.server != NULL ||
	  wsopts.user != NULL || wsopts.nick != NULL || wsopts.ip != NULL || wsopts.channel != NULL)) {
		if (parv[args] == NULL) {
			sendto_one(sptr, err_str(ERR_WHOSYNTAX), me.name, sptr->name);
			return 0;
		}
		if (strchr(parv[args], '.') != NULL) {
			wsopts.host_plus = 1;
			wsopts.host = parv[args];
		}
		else {
			wsopts.nick_plus = 1;
			wsopts.nick = parv[args];
		}
	}
	return 1;
}

static int chk_who(aClient *ac, int showall, aClient *sptr)
{
	char *host;

	if (!IsClient(ac) || (!HasMode(ac, UMODE_INVISIBLE) && showall == 0)) {
		return 0;
	}
	if (wsopts.check_umode == 1) {
		if ((wsopts.umode_plus == 1 && !((ac->umode & wsopts.umodes) == wsopts.umodes)) ||
		  (wsopts.umode_plus == 0 && ((ac->umode & wsopts.umodes) == wsopts.umodes))) {
			return 0;
		}
	}
	if (wsopts.check_away == 1) {
		if ((wsopts.away_plus == 1 && ac->user->away == NULL) || (wsopts.away_plus == 0
		  && ac->user->away != NULL)) {
			return 0;
		}
	}
	if (wsopts.serv_plus == 1) {
		if (wsopts.server != ac->uplink) {
			return 0;
		}
	}
	if (wsopts.user != NULL) {
		if ((wsopts.user_plus == 1 && uchkfn(wsopts.user, ac->username)) ||
		  (wsopts.user_plus == 0 && !uchkfn(wsopts.user, ac->username))) {
			return 0;
		}
	}
	if (wsopts.nick != NULL) {
		if ((wsopts.nick_plus == 1 && nchkfn(wsopts.nick, ac->name)) ||
		  (wsopts.nick_plus == 0 && !nchkfn(wsopts.nick, ac->name))) {
			return 0;
		}
	}
	if (wsopts.host != NULL) {
		host = HasMode(sptr, UMODE_OPER) ? ac->host : MaskedHost(ac);
		if ((wsopts.host_plus == 1 && hchkfn(wsopts.host, host)) ||
		  (wsopts.host_plus == 0 && !hchkfn(wsopts.host, host))) {
			return 0;
		}
	}
	if (wsopts.ip != NULL) {
		if ((wsopts.ip_plus == 1 && hchkfn(wsopts.ip, ac->hostip)) ||
		  (wsopts.ip_plus == 0 && !hchkfn(wsopts.ip, ac->hostip))) {
			return 0;
		}
	}
	if (wsopts.gcos != NULL) {
		if ((wsopts.gcos_plus == 1 && gchkfn(wsopts.gcos, ac->info)) ||
		  (wsopts.gcos_plus == 0 && !gchkfn(wsopts.gcos, ac->info))) {
			return 0;
		}
	}
	return 1;
}

static inline char *first_visible_channel(aClient *cptr, aClient *sptr)
{
	int secret = 0;
	aChannel *chptr = NULL;
	chanMember *cm;
	static char chnbuf[CHANNELLEN + 2];

	if (cptr->user->channel != NULL) {
		if (HasMode(sptr, UMODE_SADMIN)) {
			chptr = cptr->user->channel->chptr;
			if (!ShowChannel(sptr, chptr)) {
				secret = 1;
			}
		}
		else {
			for (cm = cptr->user->channel; cm != NULL; cm = cm->nextchan) {
				if (ShowChannel(sptr, cm->chptr)) {
					chptr = cm->chptr;
					break;
				}
			}
		}
		if (chptr != NULL) {
			if (secret == 0) {
				return chptr->chname;
			}
			ircsprintf(chnbuf, "%%%s", chptr->chname);
			return chnbuf;
		}
	}
	return "*";
}

/*
 * m_who
 *	parv[0] = sender prefix
 *	parv[1] = options
 */
int m_who(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	aClient *ac;
	chanMember *chans, *users;
	int shown = 0, i = 0, showall = HasMode(sptr, UMODE_OPER);
	char status[4];

	if (!MyClient(sptr) || !build_searchopts(sptr, parc - 1, parv + 1)) {
		return 0;
	}
	if (wsopts.gcos != NULL && !strchr(wsopts.gcos, '?') && !strchr(wsopts.gcos, '*')) {
		gchkfn = mycmp;
	}
	else {
		gchkfn = match;
	}
	if (wsopts.nick != NULL && !strchr(wsopts.nick, '?') && !strchr(wsopts.nick, '*')) {
		nchkfn = mycmp;
	}
	else {
		nchkfn = match;
	}
	if (wsopts.user != NULL && !strchr(wsopts.user, '?') && !strchr(wsopts.user, '*')) {
		uchkfn = mycmp;
	}
	else {
		uchkfn = match;
	}
	if (wsopts.host != NULL && !strchr(wsopts.host, '?') && !strchr(wsopts.host, '*')) {
		hchkfn = mycmp;
	}
	else {
		hchkfn = match;
	}
	if (wsopts.ip != NULL && !strchr(wsopts.ip, '?') && !strchr(wsopts.ip, '*')) {
		ichkfn = mycmp;
	}
	else {
		ichkfn = match;
	}
	if (wsopts.channel != NULL) {
		if (IsMember(sptr, wsopts.channel)) {
			showall = 1;
		}
		else if (SecretChannel(wsopts.channel) && HasMode(sptr, UMODE_SADMIN)) {
			showall = 1;
		}
		else if (!SecretChannel(wsopts.channel) && HasMode(sptr, UMODE_OPER)) {
			showall = 1;
		}
		else {
			showall = 0;
		}
		if (showall == 1 || !SecretChannel(wsopts.channel)) {
			for (users = wsopts.channel->members; users != NULL; users = users->nextuser) {
				ac = users->cptr;
				if (!chk_who(ac, showall, sptr)) {
					continue;
				}

				i = 0;
				status[i++] = (ac->user->away == NULL) ? 'H' : 'G';
				if (HasMode(ac, UMODE_OPER)) {
					status[i++] = '*';
				}
				else if (HasMode(ac, UMODE_INVISIBLE) && HasMode(sptr, UMODE_OPER)) {
					status[i++] = '%';
				}
				if (users->flags & CHFL_CHANOP) {
					status[i++] = '@';
				}
				else if (users->flags & CHFL_VOICE) {
					status[i++] = '+';
				}
				status[i] = '\0';

				sendto_one(sptr, rpl_str(RPL_WHOREPLY), me.name, sptr->name,
					wsopts.channel->chname, ac->username, MaskedHost(ac),
					ac->user->server, ac->name, status, WHO_HOPCOUNT(sptr, ac),
					ac->info);
			}
		}
		sendto_one(sptr, rpl_str(RPL_ENDOFWHO), me.name, sptr->name, wsopts.channel->chname);
		return 0;
	}
	else if (nchkfn == mycmp) {
		if ((ac = find_person(wsopts.nick, NULL)) != NULL) {
			if (!chk_who(ac, 1, sptr)) {
				sendto_one(sptr, rpl_str(RPL_ENDOFWHO), me.name, sptr->name,
					wsopts.host ? wsopts.host : wsopts.nick);
				return 0;
			}
			status[0] = (ac->user->away == NULL) ? 'H' : 'G';
			status[1] = (HasMode(ac, UMODE_OPER)) ? '*' : (HasMode(ac, UMODE_INVISIBLE) && HasMode(sptr, UMODE_OPER)) ? '%' : '\0';
			status[2] = '\0';
			sendto_one(sptr, rpl_str(RPL_WHOREPLY), me.name, sptr->name,
				wsopts.show_chan ? first_visible_channel(ac, sptr) : "*", ac->username,
				MaskedHost(ac), ac->user->server, ac->name, status, WHO_HOPCOUNT(sptr, ac), ac->info);
			sendto_one(sptr, rpl_str(RPL_ENDOFWHO), me.name, sptr->name,
				wsopts.host ? wsopts.host : wsopts.nick);
			return 0;
		}
		sendto_one(sptr, rpl_str(RPL_ENDOFWHO), me.name, sptr->name, wsopts.host ? wsopts.host : wsopts.nick);
		return 0;
	}
	if (lifesux > 0 && !HasMode(sptr, UMODE_OPER)) {
		sendto_one(sptr, rpl_str(RPL_LOAD2HI), me.name, sptr->name);
		return 0;
	}
	if (wsopts.search_chan == 1) {
		for (chans = sptr->user->channel; chans != NULL; chans = chans->nextchan) {
			for (users = chans->chptr->members; users != NULL; users = users->nextuser) {
				ac = users->cptr;
				if (!chk_who(ac, 1, sptr)) {
					continue;
				}
				if (shown == MAXWHOREPLIES && !HasMode(sptr, UMODE_OPER)) {
					sendto_one(sptr, err_str(ERR_WHOLIMEXCEED), me.name, sptr->name,
						MAXWHOREPLIES);
					break;
				}

				i = 0;
				status[i++] = (ac->user->away == NULL) ? 'H' : 'G';
				if (HasMode(ac, UMODE_OPER)) {
					status[i++] = '*';
				}
				else if (HasMode(ac, UMODE_INVISIBLE) && HasMode(sptr, UMODE_OPER)) {
					status[i++] = '%';
				}
				if (users->flags & CHFL_CHANOP) {
					status[i++] = '@';
				}
				else if (users->flags & CHFL_VOICE) {
					status[i++] = '+';
				}
				status[i] = '\0';

				sendto_one(sptr, rpl_str(RPL_WHOREPLY), me.name, sptr->name,
					chans->chptr->chname, ac->username, MaskedHost(ac),
					ac->user->server, ac->name, status, WHO_HOPCOUNT(sptr, ac), ac->info);
				shown++;
			}
		}
	}
	else {
		for (ac = client; ac != NULL; ac = ac->next) {
			if (!chk_who(ac, showall, sptr)) {
				continue;
			}
			if (shown == MAXWHOREPLIES && !HasMode(sptr, UMODE_OPER)) {
				sendto_one(sptr, err_str(ERR_WHOLIMEXCEED), me.name, sptr->name,
					MAXWHOREPLIES);
				break;
			}
			status[0] = (ac->user->away == NULL) ? 'H' : 'G';
			status[1] = (HasMode(ac, UMODE_OPER)) ? '*' : (HasMode(ac, UMODE_INVISIBLE) && HasMode(sptr, UMODE_OPER)) ? '%' : '\0';
			status[2] = '\0';
			sendto_one(sptr, rpl_str(RPL_WHOREPLY), me.name, sptr->name,
				wsopts.show_chan ? first_visible_channel(ac, sptr) : "*",
				ac->username, MaskedHost(ac), ac->user->server, ac->name, status,
				WHO_HOPCOUNT(sptr, ac), ac->info);
			shown++;
		}
	}
	sendto_one(sptr, rpl_str(RPL_ENDOFWHO), me.name, sptr->name,
		wsopts.host != NULL ? wsopts.host : wsopts.nick != NULL ? wsopts.nick :
		wsopts.user != NULL ? wsopts.user : wsopts.gcos != NULL ? wsopts.gcos :
		wsopts.server != NULL ? wsopts.server->name : wsopts.ip != NULL ? wsopts.ip : "*");
	return 0;
}
