/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_zline.c: ZLINE command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_zline.c,v 1.16 2002/12/15 21:26:20 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

#ifndef STATIC_MODULES
Module _module_header = {
	"modules/m_zline.so",
	"/ZLINE command",
	4,
	"$Revision: 1.16 $",
	0, NULL, NULL
};
void _module_load() {
	addCommand("ZLINE", m_zline, MAXPARA, MSG_NONE);
	return;
}
void _module_unload() {
	delCommand("ZLINE", m_zline);
	return;
}
#endif

/*
 * m_zline
 *	parv[0] = sender prefix
 *	parv[1] = ip address
 *	parv[2] = reason
 */
int m_zline(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	ConfigItem_ban *ban;
	aClient *acptr;
	char *host, *reason, *p;
	int dots = 0;

	if (!OPHasFlag(sptr, OFLAG_ZAP)) {
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		return 0;
	}
	if (parc < 2 || BadPtr(parv[1])) {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "ZLINE");
		return 0;
	}
	if (strchr(parv[1], '@')) {
		sendto_one(sptr, ":%s NOTICE %s :*** Please specify an IP address or client to Z:Line",
			me.name, parv[0]);
		return 0;
	}

	host = parv[1];
	for (p = host; *p; p++) {
		if (*p == '.') {
			dots++;
		}
		else if (*p == '*') {
			if (dots != 3) {
				sendto_one(sptr, ":%s NOTICE %s :*** Z:Line format error - please use "
					"xxx.xxx.xxx.xxx or xxx.xxx.xxx.*", me.name, parv[0]);
				return 0;
			}
			else {
				*(p + 1) = '\0';
				break;
			}
		}
		else {
			if (!isdigit(*p)) {
				if ((acptr = find_chasing(sptr, parv[1], NULL)) == NULL) {
					return 0;
				}
				if (acptr->user == NULL || IsServer(acptr)) {
					return 0;
				}
				if (!MyConnect(acptr)) {
					sendto_one(sptr, ":%s NOTICE %s :*** You cannot Z:Line remote "
						"clients", me.name, parv[0]);
					return 0;
				}
				host = acptr->hostip;
				break;
			}
		}
	}

	if (parc > 2 && !BadPtr(parv[2])) {
		reason = parv[2];
	}
	else {
		reason = "no reason";
	}

	for (ban = ci_ban; ban != NULL; ban = (ConfigItem_ban *)ban->next) {
		if (ban->type & BAN_IP) {
			if (!match(ban->mask, host)) {
				sendto_one(sptr, ":%s NOTICE %s :*** %s is already on the temporary Z:Line list as %s",
					me.name, parv[0], host, ban->mask);
				return 0;
			}
		}
	}
	ban = (ConfigItem_ban *)MyMallocX(sizeof(ConfigItem_ban));
	ban->type = BAN_IP;
	DupString(ban->mask, host);
	DupString(ban->reason, reason);
	ban->flag.temporary = 1;
	add_ConfigItem((ConfigItem *)ban, (ConfigItem **)&ci_ban);

	ircdlog(LOG_BAN, "TEMPZLINE for %s (%s) added by %s", ban->mask, ban->reason, get_client_name(sptr, TRUE));
	sendto_realops("%s (%s@%s) added temporary Z:Line for *@%s (%s)", parv[0], sptr->username,
		MaskedHost(sptr), ban->mask, ban->reason);
	do_bancheck = 1;
	return 0;
}
