/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/adns.c: libadns interface functions
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: adns.c,v 1.9 2002/11/10 19:56:04 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "res.h"
#include "numeric.h"
#include "h.h"
#include "memory.h"
#include <stdlib.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/socket.h>
#include "nameser.h"
#include "inet.h"
#include "internal.h"
#include "adns.h"

adns_state ircd_dns;

static struct timeval ircd_dnstime;

void del_dns_queries(DNSQuery *q)
{
	if (q != NULL && q->query != NULL) {
		adns_cancel(q->query);
	}
	return;
}

int init_resolver()
{
	int i;

	gettimeofday(&ircd_dnstime, NULL);
	i = adns_init(&ircd_dns, adns_if_noautosys, 0);
	if (ircd_dns == NULL) {
		return 0;
	}
	addEvent("dns_timeout", dns_timeout, NULL, 2, 1);
	return 1;
}

void restart_resolver()
{
	adns__rereadconfig(ircd_dns);
	return;
}

void dns_timeout()
{
	gettimeofday(&ircd_dnstime, NULL);
	adns_processtimeouts(ircd_dns, &ircd_dnstime);
	return;
}

void do_dns_callbacks()
{
	adns_query q, r;
	adns_answer *answer;
	DNSQuery *query;

	adns_forallqueries_begin(ircd_dns);
	while ((q = adns_forallqueries_next(ircd_dns, (void **)&r)) != NULL) {
		switch (adns_check(ircd_dns, &q, &answer, (void **)&query)) {
			case 0:
				if (query->callback == NULL) {
					ircdlog(LOG_ERROR, "FATAL: dns_do_callbacks() got NULL callback ptr at "
						"check 0, query %lX", query);
					abort();
				}
				query->query = NULL;
				query->callback(query->ptr, answer);
				MyFree(query);
				break;
			case EAGAIN:
				break;
			default:
				if (query->callback == NULL) {
					ircdlog(LOG_ERROR, "FATAL: dns_do_callbacks() got NULL callback ptr at "
						"check default, query %lX", query);
					abort();
				}
				query->query = NULL;
				query->callback(query->ptr, NULL);
				MyFree(query);
				break;
		}
	}
	return;
}

static void do_dns_fd_io()
{
#ifdef HAVE_POLL_ON
	struct pollfd pfd[MAXFD_POLL];
	int nfds = MAXFD_POLL;
	int timeout = 0;

	adns_beforepoll(ircd_dns, pfd, &nfds, &timeout, &ircd_dnstime);
	poll(pfd, nfds, timeout);
	adns_afterpoll(ircd_dns, pfd, nfds, &ircd_dnstime);
#else
	struct timeval tv;
	int maxfd = 2;
	fd_set readfds, writefds, exceptfds;

	tv.tv_sec = 0;
	tv.tv_usec = 0;
	FD_ZERO(&readfds);
	FD_ZERO(&writefds);
	FD_ZERO(&exceptfds);

	adns_beforeselect(ircd_dns, &maxfd, &readfds, &writefds, &exceptfds, 0, &tv, 0);
	select(maxfd, &readfds, &writefds, &exceptfds, &tv);
	adns_afterselect(ircd_dns, maxfd, &readfds, &writefds, &exceptfds, &tv);
#endif
	return;
}

void process_dns_queries()
{
	gettimeofday(&ircd_dnstime, NULL);
	do_dns_fd_io();
	adns_processany(ircd_dns);
	do_dns_callbacks();
	dns_timeout();
	return;
}

void dns_gethost(const char *name, DNSQuery *q)
{
	if (ircd_dns->nservers > 0) {
		adns_submit(ircd_dns, name, adns_r_addr, adns_qf_owner, q, &q->query);
	}
	return;
}

void dns_getaddr(struct in_addr *addr, DNSQuery *q)
{
	struct sockaddr_in sin;
	if (ircd_dns->nservers > 0) {
		sin.sin_family = AF_INET;
		sin.sin_port = 0;
		sin.sin_addr.s_addr = addr->s_addr;
		adns_submit_reverse(ircd_dns, (struct sockaddr *)&sin, adns_r_ptr,
			adns_qf_owner|adns_qf_cname_loose|adns_qf_quoteok_anshost, q, &q->query);
	}
	return;
}

void report_dns_nameservers(aClient *cptr)
{
	int i;

	sendto_one(cptr, ":%s NOTICE %s :*** Current DNS Nameservers:", me.name, cptr->name);
	for (i = 0; i < ircd_dns->nservers; i++) {
		sendto_one(cptr, ":%s NOTICE %s :*** %s", me.name, cptr->name,
			inetntoa((char *)&ircd_dns->servers[i].addr));
	}
	return;
}
