/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/auth.c: password authentication engine
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: auth.c,v 1.19 2002/11/19 19:31:12 rageircd Exp $
 */

#include "conf2.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "patchlevel.h"
#include "ssl.h"
#include "config.h"
#include "memory.h"
#include "send.h"
#include "h.h"
#include <time.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

static const char enc_chars[] = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789./";

static Auth AuthTypes[] = {
#ifdef HAVE_OPENSSL_ON
	{ "ripemd/160", AUTH_RIPEMD_160 },
	{ "sha1/160", AUTH_SHA1_160 },
	{ "md5/128", AUTH_MD5_128 },
#endif
	{ "des/56", AUTH_DES_56 },
	{ "md5/56", AUTH_MD5_56 },
	{ "plaintext", AUTH_PLAINTEXT },
	{ NULL, 0 }
};

static unsigned int auth_str_to_type(char *);
static char *source_parse_userhost(char *);

extern int base64_encode(const char *, size_t, char *, size_t);
extern char *crypt(const char *, const char *);

Auth *die_auth;
Auth *restart_auth;

static unsigned int auth_str_to_type(char *str)
{
	Auth *auth;
	for (auth = AuthTypes; auth->string != NULL; auth++) {
		if (!mycmp(auth->string, str)) {
			return auth->type;
		}
	}
	return -1;
}

Auth *auth_parse_conf_block(ConfigEntry *ce, char *directive)
{
	unsigned int type = -1;
	char *string = NULL;
	ConfigEntry *cep;
	Auth *auth;

	if (ce->ce_entries == NULL) {
		report(0, "%s:%d: %s: empty block", ce->ce_fileptr->cf_filename, ce->ce_varlinenum, directive);
		return NULL;
	}
	for (cep = ce->ce_entries; cep != NULL; cep = cep->ce_next) {
		CheckVar(cep, directive);
		if (!mycmp(cep->ce_varname, "string")) {
			if (BadPtr(cep->ce_vardata)) {
				report(0, "%s:%d: %s::string: missing variable parameter",
					cep->ce_fileptr->cf_filename, cep->ce_varlinenum, directive);
				continue;
			}
			string = cep->ce_vardata;
		}
		else if (!mycmp(cep->ce_varname, "type")) {
			if (BadPtr(cep->ce_vardata)) {
				report(0, "%s:%d: %s::type: missing variable parameter",
					cep->ce_fileptr->cf_filename, cep->ce_varlinenum, directive);
				continue;
			}
			if ((type = auth_str_to_type(cep->ce_vardata)) == -1) {
				report(0, "%s:%d: %s::type: unknown type %s", cep->ce_fileptr->cf_filename,
					cep->ce_varlinenum, directive, cep->ce_vardata);
				continue;
			}
		}
		else {
			UnknownVar(cep, directive);
		}
	}
	if (string == NULL) {
		report(0, "%s:%d: %s::string: missing parameter, removing section",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum, directive);
		return NULL;
	}
	if (type == -1) {
		report(0, "%s:%d: %s::type: missing parameter, removing section",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum, directive);
		return NULL;
	}
	auth = (Auth *)MyMallocX(sizeof(Auth));
	DupString(auth->string, string);
	auth->type = type;
	return auth;
}

int auth_check(Auth *auth, char *passwd)
{
#ifdef HAVE_OPENSSL_ON
	static char buf[512];
#endif
	char salt[12], *encr;
	int i = 0;

	if (auth == NULL || passwd == NULL) {
		return 0;
	}
	switch (auth->type) {
#ifdef HAVE_OPENSSL_ON
		case AUTH_RIPEMD_160:
			if ((i = base64_encode(RIPEMD160(passwd, strlen(passwd), NULL),
			  RIPEMD160_DIGEST_LENGTH, buf, 512)) > 0) {
				if (!irccmp(buf, auth->string)) {
					return 1;
				}
			}
			break;
		case AUTH_SHA1_160:
			if ((i = base64_encode(SHA1(passwd, strlen(passwd), NULL),
			  SHA_DIGEST_LENGTH, buf, 512)) > 0) {
				if (!irccmp(buf, auth->string)) {
					return 1;
				}
			}
			break;
		case AUTH_MD5_128:
			if ((i = base64_encode(MD5(passwd, strlen(passwd), NULL),
			  MD5_DIGEST_LENGTH, buf, 512)) > 0) {
				if (!irccmp(buf, auth->string)) {
					return 1;
				}
			}
			break;
#endif
		case AUTH_MD5_56:
			salt[0] = '$';
			salt[1] = '1';
			salt[2] = '$';
			for (i = 0; i < 8; i++) {
				if (auth->string[i + 3] == '$') {
					break;
				}
				salt[i + 3] = auth->string[i + 3];
			}
			salt[i + 3] = '\0';
			encr = crypt(passwd, salt);

			if (!irccmp(encr, auth->string)) {
				return 1;
			}
			break;
		case AUTH_DES_56:
			if (strlen(passwd) >= 2) {
				for (i = 0; i < 2; i++) {
					salt[i] = auth->string[i];
				}
				salt[i] = '\0';
				encr = crypt(passwd, salt);
				if (!irccmp(encr, auth->string)) {
					return 1;
				}
			}
			break;
		case AUTH_PLAINTEXT:
			if (!irccmp(passwd, auth->string)) {
				return 1;
			}
			break;
	}
	return 0;
}

void auth_destroy_struct(Auth *auth)
{
	if (auth != NULL) {
		if (auth->string != NULL) {
			MyFree(auth->string);
		}
		MyFree(auth);
	}
	return;
}

void auth_handle_mkpasswd(aClient *cptr, char *type_str, char *passwd)
{
	unsigned int type = -1;
#ifdef HAVE_OPENSSL_ON
	static char buf[512];
#endif
	char salt[12];
	int i = 0;

	if ((type = auth_str_to_type(type_str)) == -1 || type == AUTH_PLAINTEXT) {
		sendto_one(cptr, ":%s NOTICE %s :*** %s is an unknown encryption method.",
			me.name, cptr->name, type_str);
		return;
	}

	srandom(time(NULL));
	switch (type) {
#ifdef HAVE_OPENSSL_ON
		case AUTH_RIPEMD_160:
			if ((i = base64_encode(RIPEMD160(passwd, strlen(passwd), NULL),
			  RIPEMD160_DIGEST_LENGTH, buf, 512)) > 0) {
				sendto_one(cptr, ":%s NOTICE %s :*** Plaintext: %s Encrypted (ripemd/160): %s",
					me.name, cptr->name, passwd, buf);
				return;
			}
			break;
		case AUTH_SHA1_160:
			if ((i = base64_encode(SHA(passwd, strlen(passwd), NULL),
			  SHA_DIGEST_LENGTH, buf, 512)) > 0) {
				sendto_one(cptr, ":%s NOTICE %s :*** Plaintext: %s Encrypted (sha1/160): %s",
					me.name, cptr->name, passwd, buf);
				return;
			}
			break;
		case AUTH_MD5_128:
			if ((i = base64_encode(MD5(passwd, strlen(passwd), NULL),
			  MD5_DIGEST_LENGTH, buf, 512)) > 0) {
				sendto_one(cptr, ":%s NOTICE %s :*** Plaintext: %s Encrypted (md5/128): %s",
					me.name, cptr->name, passwd, buf);
				return;
			}
			break;
#endif
		case AUTH_DES_56:
			for (i = 0; i < 2; i++) {
				salt[i] = enc_chars[random() % 64];
			}
			salt[i] = '\0';
			sendto_one(cptr, ":%s NOTICE %s :*** Plaintext: %s Encrypted (des/56): %s",
				me.name, cptr->name, passwd, crypt(passwd, salt));
			return;
			break;
		case AUTH_MD5_56:
			salt[0] = '$';
			salt[1] = '1';
			salt[2] = '$';
			for (i = 0; i < 8; i++) {
				salt[i + 3] = enc_chars[random() % 64];
			}
			salt[i + 3] = '\0';
			sendto_one(cptr, ":%s NOTICE %s :*** Plaintext: %s Encrypted (md5/56): %s",
				me.name, cptr->name, passwd, crypt(passwd, salt));
			return;
			break;
		default:
			ircdlog(LOG_ERROR, "auth_handle_mkpasswd() encountered unknown encryption %d",
				(int)type);
	}
	return;
}

static char *source_parse_userhost(char *userhost)
{
	char *p;
	for (p = userhost; *p != '\0'; p++) {
		if (*p == '@') {
			break;
		}
	}
	if (*p == '\0' || (*p != '\0' && *userhost == '@')) {
		static char ret[HOSTLEN + 3];
		ircsprintf(ret, "*%s%s", (*p != '\0' && *userhost == '@') ? "" : "@", userhost);
		return ret;
	}
	return userhost;
}

void *source_parse_conf_block(ConfigEntry *ce, char *directive, Source **src_list)
{
	ConfigEntry *cep;
	Source *src;

	CheckBlockEntries(ce, directive);
	for (cep = ce->ce_entries; cep != NULL; cep = cep->ce_next) {
		CheckVar(cep, directive);
		src = (Source *)MyMallocX(sizeof(Source));
		ConfDupString(src->userhost, source_parse_userhost(cep->ce_varname));
		src->prev = NULL;
		if ((src->next = *src_list) != NULL) {
			src->next->prev = src;
		}
		*src_list = src;
	}
	return NULL;
}

Source *source_del(Source *src_item, Source **src_list)
{
	Source *src, *next = NULL;
	for (src = *src_list; src != NULL; src = src->next) {
		if (src == src_item) {
			next = src->next;
			if (src->prev != NULL) {
				src->prev->next = src->next;
			}
			else {
				*src_list = src->next;
			}
			if (src->next != NULL) {
				src->next->prev = src->prev;
			}
			return next;
		}
	}
	return NULL;
}

void source_destroy_list(Source **src_list)
{
	Source *src, s;
	for (src = *src_list; src != NULL; src = src->next) {
		s.next = source_del(src, src_list);
		MyFree(src->userhost);
		MyFree(src);
		src = (Source *)&s;
	}
	return;
}

int source_check(Source **src_list, char *userhost)
{
	Source *src;
	for (src = *src_list; src != NULL; src = src->next) {
		if (!match(src->userhost, userhost)) {
			return 1;
		}
	}
	return 0;
}
