/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/blalloc.c: memory block allocator
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: blalloc.c,v 1.11 2002/11/19 23:16:51 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#include "numeric.h"
#include "blalloc.h"
#include "setup.h"
#include "memory.h"

#ifdef HAVE_BLALLOC_ON

void blockheap_garbage_collect()
{
	BlockHeapGarbageCollect(channel_heap);
	BlockHeapGarbageCollect(chanmember_heap);
	BlockHeapGarbageCollect(channelban_heap);
	BlockHeapGarbageCollect(local_client_heap);
	BlockHeapGarbageCollect(remote_client_heap);
	BlockHeapGarbageCollect(local_user_heap);
	BlockHeapGarbageCollect(remote_user_heap);
	BlockHeapGarbageCollect(slink_heap);
#ifdef FLUD
	BlockHeapGarbageCollect(fludbot_heap);
#endif
#ifdef HAVE_THROTTLE_ON
	BlockHeapGarbageCollect(throttle_heap);
	BlockHeapGarbageCollect(hashent_heap);
#endif
	return;
}

void init_blockheap()
{
	addEvent("blockheap_garbage_collect", blockheap_garbage_collect, NULL, BLOCKHEAP_GARBAGECOLLECT_FREQ, 1);
	return;
}

static int newblock(BlockHeap *bh)
{
	Block *b = NULL;
	int i;

	if ((b = (Block *)MyMalloc(sizeof(Block))) == NULL) {
		ircdlog(LOG_ERROR, "newblock failed (allocation failed)");
		return 1;
	}
	b->freeElems = bh->elemsPerBlock;
	b->next = bh->base;

	if ((b->allocMap = (long *)MyMalloc(sizeof(long) * (bh->numlongs + 1))) == NULL) {
		ircdlog(LOG_ERROR, "newblock failed (b->allocMap allocation failed)");
		MyFree(b);
		return 1;
	}
	memset((void *)b->allocMap, '\0', sizeof(long) * (bh->numlongs + 1));

	if ((b->elems = (void *)MyMalloc((bh->elemsPerBlock + 1) * bh->elemSize)) == NULL) {
		ircdlog(LOG_ERROR, "newblock failed (b->elems allocation failed)");
		MyFree(b->allocMap);
		MyFree(b);
		return 1;
	}
	b->endElem = (void *)((unsigned long)b->elems + (unsigned long)((bh->elemsPerBlock - 1) * bh->elemSize));

	for (i = 0; i < bh->numlongs; i++) {
		b->allocMap[i] = 0L;
	}

	bh->blocksAllocated++;
	bh->base = b;
	return 0;
}

BlockHeap *BlockHeapCreate(size_t elemsize, int elemsperblock)
{
	BlockHeap *bh;

	if (elemsize < 0 || elemsperblock < 0) {
		ircdlog(LOG_ERROR, "FATAL: BlockHeapCreate failed (elemsize < 0 || elemsperblock < 0)");
		outofmemory();
	}

	if ((bh = (BlockHeap *)MyMalloc(sizeof(BlockHeap))) == NULL) {
		ircdlog(LOG_ERROR, "FATAL: BlockHeapCreate failed (allocation failed)");
		outofmemory();
	}
	elemsize = elemsize + (elemsize & (sizeof(void *) - 1));
	bh->elemSize = elemsize;
	bh->elemsPerBlock = elemsperblock;
	bh->blocksAllocated = 0;
	bh->freeElems = 0;
	bh->numlongs = (bh->elemsPerBlock / (sizeof(long) * 8)) + 1;
	if (bh->elemsPerBlock % (sizeof(long) * 8) == 0) {
		bh->numlongs--;
	}
	bh->base = NULL;

	if (newblock(bh) > 0) {
		MyFree(bh);
		ircdlog(LOG_ERROR, "FATAL: newblock in BlockHeapCreate failed");
		outofmemory();
	}
	if (bh == NULL) {
		ircdlog(LOG_ERROR, "FATAL: BlockHeapCreate failed (bh is unallocated?!)");
		outofmemory();
	}
	return bh;
}

void *BlockHeapALLOC(BlockHeap *bh)
{
	Block *b;
	int unit;
	unsigned long mask, ctr;

	if (bh == NULL) {
		ircdlog(LOG_ERROR, "BlockHeapAlloc failed (block is NULL)");
		return (void *)NULL;
	}
	if (bh->freeElems == 0) {
		if (newblock(bh) > 0) {
			ircdlog(LOG_ERROR, "newblock failed in BlockHeapAlloc");
			return NULL;
		}
		else {
			b = bh->base;
			b->allocMap[0] = 0x1L;
			b->freeElems--;
			bh->freeElems--;
			if (bh->base->elems == NULL) {
				ircdlog(LOG_ERROR, "BlockHeapAlloc failed (bh->base->elems is NULL)");
				return (void *)NULL;
			}
		}
	}

	for (b = bh->base; b != NULL; b = b->next) {
		if (b->freeElems > 0) {
			mask = 0x1L;
			ctr = 0;
			unit = 0;

			while (unit < bh->numlongs) {
				if (mask == 0x1L && b->allocMap[unit] == ~0) {
					unit++;
					ctr = 0;
					continue;
				}

				if (!(mask & b->allocMap[unit])) {
					b->allocMap[unit] |= mask;
					b->freeElems--;
					bh->freeElems--;
					return ((void *)((unsigned long)b->elems + ((unit * sizeof(unsigned long) *
					  8 + ctr) * (unsigned long)bh->elemSize)));
				}

				mask <<= 1;
				ctr++;
				if (mask == 0) {
					mask = 0x1L;
					unit++;
					ctr = 0;
				}
			}
		}
	}
	ircdlog(LOG_ERROR, "BlockHeapAlloc failed (end of function)");
	return (void *)NULL;
}

int BlockHeapFREE(BlockHeap *bh, void *ptr)
{
	Block *b;
	unsigned long ctr;
	unsigned long mask;

	if (bh == NULL) {
		ircdlog(LOG_ERROR, "BlockHeapFree failed (block is NULL)");
		return 1;
	}
	for (b = bh->base; b != NULL; b = b->next) {
		if (ptr >= b->elems && ptr <= b->endElem) {
			ctr = ((unsigned long)ptr - (unsigned long)b->elems) / (unsigned long)bh->elemSize;
			mask = 1L << (ctr % (sizeof(long) * 8));
			ctr = ctr / (sizeof(long) * 8);

			if ((b->allocMap[ctr] & mask) == 0) {
				ircdlog(LOG_ERROR, "BlockHeapFree failed (bit already clear in map)");
			}
			else {
				b->allocMap[ctr] = b->allocMap[ctr] & ~mask;
				b->freeElems++;
				bh->freeElems++;
			}
			return 0;
		}
	}
	ircdlog(LOG_ERROR, "BlockHeapFree failed (end of function)");
	return 1;
}

int BlockHeapGarbageCollect(BlockHeap *bh)
{
	Block *b, *last = NULL;
	int i;

	if (bh == NULL) {
		ircdlog(LOG_ERROR, "BlockHeapGarbageCollect failed (block is NULL)");
		return 1;
	}
	if (bh->freeElems < bh->elemsPerBlock) {
		return 0;
	}
	for (b = bh->base; b != NULL;) {
		for (i = 0; i < bh->numlongs; i++) {
			if (b->allocMap[i]) {
				break;
			}
		}
		if (i == bh->numlongs) {
			MyFree(b->elems);
			MyFree(b->allocMap);

			if (last != NULL) {
				last->next = b->next;
				MyFree(b);
				b = last->next;
			}
			else {
				bh->base = b->next;
				MyFree(b);
				b = bh->base;
			}
			bh->blocksAllocated--;
			bh->freeElems -= bh->elemsPerBlock;
		}
		else {
			last = b;
			b = b->next;
		}
	}
	return 0;
}

#endif /* HAVE_BLALLOC_ON */
