/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/channel_mode.c: channel mode handling
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: channel_mode.c,v 1.9 2002/11/19 23:47:13 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "channel.h"
#include "h.h"
#include "memory.h"

char chmode_string[32];

static char buf[BUFSIZE];
static char modebuf[REALMODEBUFLEN];
static char parabuf[REALMODEBUFLEN];
static int used_samode = 0;

aCtab cFlagTab[] = {
	{ MODE_BAN, 'b', 0, 1 },
	{ MODE_INVITEONLY, 'i', 1, 0 },
	{ MODE_KEY, 'k', 0, 1 },
	{ MODE_LIMIT, 'l', 0, 1 },
	{ MODE_MODERATED, 'm', 0, 0 },
	{ MODE_NOPRIVMSGS, 'n', 1, 0 },
	{ MODE_CHANOP, 'o', 0, 1 },
	{ MODE_PRIVATE, 'p', 1, 0 },
	{ MODE_REGISTERED, 'r', 0, 0 },
	{ MODE_REGONLY, 'R', 0, 0 },
	{ MODE_SECRET, 's', 1, 0 },
	{ MODE_TOPICLIMIT, 't', 1, 0 },
	{ MODE_VOICE, 'v', 1, 1 },
	{ MODE_NOCOLOUR, 'c', 1, 0 },
	{ MODE_OPERONLY, 'O', 0, 0 },
	{ MODE_ADMINONLY, 'A', 0, 0 },
	{ MODE_HALFOP, 'h', 1, 1 },
	{ MODE_CHANADMIN, 'a', 0, 1 },
	{ MODE_STRIPCOLOUR, 'S', 1, 0 },
	{ MODE_EXBAN, 'e', 0, 1 },
	{ MODE_NOCTCP, 'C', 1, 0 },
	{ MODE_NONICKCH, 'N', 1, 0 },
	{ MODE_INVEX, 'I', 0, 1 },
	{ 0, 0, 0, 0 }
};

static int check_halfop(u_int modetype)
{
	aCtab *tab;
	for (tab = &cFlagTab[0]; tab->mode != 0; tab++) {
		if (tab->mode == modetype) {
			return (tab->halfop == 1) ? TRUE : FALSE;
		}
	}
	return TRUE;
}

static int check_chankey(char *param)
{
	char *p;
	for (p = param; *p != '\0'; p++) {
		if (*p == ' ' || *p == '*' || *p == ':' || *p == ',') {
			return 0;
		}
	}
	return 1;
}

void build_chmode_string(void)
{
	char *p = &chmode_string[0];
	aCtab *tab;

	for (tab = &cFlagTab[0]; tab->mode != 0; tab++) {
		*p++ = tab->flag;
	}
	*p++ = ' ';
	for (tab = &cFlagTab[0]; tab->mode != 0; tab++) {
		if (tab->param == 1) {
			*p++ = tab->flag;
		}
	}
	*p = '\0';
	return;
}

int is_chan_admin(aClient *cptr, aChannel *chptr)
{
	chanMember *cm;
	if (chptr != NULL) {
		if ((cm = find_user_member(cptr->user->channel, chptr)) != NULL) {
			return (cm->flags & CHFL_CHANADMIN) ? 1 : 0;
		}
	}
	return 0;
}

int is_chan_op(aClient *cptr, aChannel *chptr)
{
	chanMember *cm;
	if (chptr != NULL) {
		if ((cm = find_user_member(cptr->user->channel, chptr)) != NULL) {
			return (cm->flags & CHFL_CHANOP) ? 1 : 0;
		}
	}
	return 0;
}

int is_half_op(aClient *cptr, aChannel *chptr)
{
	chanMember *cm;
	if (chptr != NULL) {
		if ((cm = find_user_member(cptr->user->channel, chptr)) != NULL) {
			return (cm->flags & CHFL_HALFOP) ? 1 : 0;
		}
	}
	return 0;
}

int has_voice(aClient *cptr, aChannel *chptr)
{
	chanMember *cm;
	if (chptr != NULL) {
		if ((cm = find_user_member(cptr->user->channel, chptr)) != NULL) {
			return (cm->flags & CHFL_VOICE) ? 1 : 0;
		}
	}
	return 0;
}

void channel_modes(aClient *cptr, char *mbuf, char *pbuf, aChannel *chptr)
{
	aCtab *tab;
	char *ptr;

	*mbuf++ = '+';
	for (tab = &cFlagTab[0]; tab->mode != 0; tab++) {
		if (tab->param == 1 && tab->mode != MODE_KEY && tab->mode != MODE_LIMIT) {
			continue;
		}
		if (chptr->mode.mode & tab->mode) {
			*mbuf++ = tab->flag;
		}
	}
	if (chptr->mode.limit != 0) {
		*mbuf++ = 'l';
		if (IsMember(cptr, chptr) || IsServer(cptr) || IsULine(cptr)) {
			for (ptr = my_itoa(chptr->mode.limit); *ptr != '\0'; ptr++) {
				*pbuf++ = *ptr;
			}
		}
	}
	if (*chptr->mode.key != '\0') {
		*mbuf++ = 'k';
		if (IsMember(cptr, chptr) || IsServer(cptr) || IsULine(cptr)) {
			if (chptr->mode.limit != 0) {
				*pbuf++ = ' ';
			}
			for (ptr = chptr->mode.key; *ptr != '\0'; ptr++) {
				*pbuf++ = *ptr;
			}
		}
	}
	*mbuf = '\0';
	*pbuf = '\0';
	return;
}

static void make_sjoin_buf(aClient *cptr, aChannel *chptr, char *modebuf, char *parabuf)
{
	if (HasCapab(cptr, CAPAB_SSJv2)) {
		ircsprintf(buf, ":%s SJOIN %ld %s %s %s :", me.name, chptr->channelts, chptr->chname,
			modebuf, *parabuf == '\0' ? "<none>" : parabuf);
	}
	else {
		if (*parabuf == '\0') {
			ircsprintf(buf, ":%s SJOIN %ld %s %s :", me.name, chptr->channelts, chptr->chname,
				modebuf);
		}
		else {
			ircsprintf(buf, ":%s SJOIN %ld %s %s %s :", me.name, chptr->channelts, chptr->chname,
				modebuf, parabuf);
		}
	}
	return;
}

void send_channel_modes(aClient *cptr, aChannel *chptr)
{
	chanMember *cm;
	char *bufptr;
	aBan *ban;
	int n = 0;

	if ((!HasCapab(cptr, CAPAB_SSJOIN) && !HasCapab(cptr, CAPAB_SSJv2)) || *chptr->chname != '#') {
		return;
	}

	modebuf[0] = '\0';
	parabuf[0] = '\0';
	channel_modes(cptr, modebuf, parabuf, chptr);
	make_sjoin_buf(cptr, chptr, modebuf, parabuf);
	bufptr = buf + strlen(buf);

	for (cm = chptr->members; cm != NULL; cm = cm->nextuser) {
		if (HasCapab(cptr, CAPAB_SSJv2) && cm->flags & CHFL_CHANADMIN) {
			*bufptr++ = '*';
		}
		if (cm->flags & CHFL_CHANOP) {
			*bufptr++ = '@';
		}
		if (HasCapab(cptr, CAPAB_SSJv2) && cm->flags & CHFL_HALFOP) {
			*bufptr++ = '%';
		}
		if (cm->flags & CHFL_VOICE) {
			*bufptr++ = '+';
		}
		strcpy(bufptr, cm->cptr->name);
		bufptr += strlen(bufptr);
		*bufptr++ = ' ';
		n++;

		if (bufptr - buf > BUFSIZE - 80) {
			*bufptr++ = '\0';
			if (bufptr[-1] == ' ') {
				bufptr[-1] = '\0';
			}
			sendto_one(cptr, "%s", buf);
			make_sjoin_buf(cptr, chptr, modebuf, parabuf);
			n = 0;
			bufptr = buf + strlen(buf);
		}
	}
	if (n > 0) {
		*bufptr++ = '\0';
		if (bufptr[-1] == ' ') {
			bufptr[-1] = '\0';
		}
		sendto_one(cptr, "%s", buf);
		make_sjoin_buf(cptr, chptr, modebuf, parabuf);
		n = 0;
		bufptr = buf + strlen(buf);
	}

	if (HasCapab(cptr, CAPAB_SSJv2)) {
		for (ban = chptr->banlist; ban != NULL; ban = ban->next) {
			*bufptr++ = '&';
			strcpy(bufptr, ban->banstr);
			bufptr += strlen(bufptr);
			*bufptr++ = ' ';
			n++;

			if (bufptr - buf > BUFSIZE - 80) {
				*bufptr++ = '\0';
				if (bufptr[-1] == ' ') {
					bufptr[-1] = '\0';
				}
				sendto_one(cptr, "%s", buf);
				make_sjoin_buf(cptr, chptr, modebuf, parabuf);
				n = 0;
				bufptr = buf + strlen(buf);
			}
		}
		for (ban = chptr->exbanlist; ban != NULL; ban = ban->next) {
			*bufptr++ = '"';
			strcpy(bufptr, ban->banstr);
			bufptr += strlen(bufptr);
			*bufptr++ = ' ';
			n++;

			if (bufptr - buf > BUFSIZE - 80) {
				*bufptr++ = '\0';
				if (bufptr[-1] == ' ') {
					bufptr[-1] = '\0';
				}
				sendto_one(cptr, "%s", buf);
				make_sjoin_buf(cptr, chptr, modebuf, parabuf);
				n = 0;
				bufptr = buf + strlen(buf);
			}
		}
		for (ban = chptr->invexlist; ban != NULL; ban = ban->next) {
			*bufptr++ = '$';
			strcpy(bufptr, ban->banstr);
			bufptr += strlen(bufptr);
			*bufptr++ = ' ';
			n++;

			if (bufptr - buf > BUFSIZE - 80) {
				*bufptr++ = '\0';
				if (bufptr[-1] == ' ') {
					bufptr[-1] = '\0';
				}
				sendto_one(cptr, "%s", buf);
				make_sjoin_buf(cptr, chptr, modebuf, parabuf);
				n = 0;
				bufptr = buf + strlen(buf);
			}
		}
		if (n > 0) {
			*bufptr++ = '\0';
			if (bufptr[-1] == ' ') {
				bufptr[-1] = '\0';
			}
			sendto_one(cptr, "%s", buf);
		}
	}
	return;
}

int do_mode_char(aChannel *chptr, u_int modetype, char modechar, char *param, u_int what, aClient *cptr,
	u_int *pcount, char pvar[MAXMODEPARAMS][MODEBUFLEN + 3], char bounce)
{
	char tmpbuf[256], *tmpstr, *modename = "?", modeflag = '?';
	char nuhbuf[NICKLEN + USERLEN + HOSTLEN + 6];
	int retval = 0, chasing = 0, x;
	unsigned int tmp = 0;
	chanMember *member;
	aClient *who;
	aCtab *tab;

	if (!IsULine(cptr) && !IsServer(cptr) && used_samode == 1) {
		if (!is_chan_op(cptr, chptr) && is_half_op(cptr, chptr)) {
			if (check_halfop(modetype) == FALSE) {
				for (tab = &cFlagTab[0]; tab->mode != 0; tab++) {
					if (tab->mode == modetype) {
						sendto_one(cptr, err_str(ERR_ONLYHALFOP), me.name, cptr->name, tab->flag);
					}
				}
				return 0;
			}
		}
	}
	switch (modetype) {
		case MODE_REGISTERED:
			if (!IsServer(cptr)) {
				break;
			}
		case MODE_ADMINONLY:
			if (!HasMode(cptr, UMODE_NETADMIN|UMODE_ADMIN|UMODE_COADMIN)) {
				sendto_one(cptr, err_str(ERR_NOPRIVILEGES), me.name, cptr->name);
				break;
			}
			if (chptr->mode.mode & MODE_OPERONLY) {
				chptr->mode.mode &= ~MODE_OPERONLY;
			}
		case MODE_OPERONLY:
			if (!HasMode(cptr, UMODE_OPER)) {
				sendto_one(cptr, err_str(ERR_NOPRIVILEGES), me.name, cptr->name);
				break;
			}
			if (chptr->mode.mode & MODE_ADMINONLY) {
				sendto_one(cptr, ":%s NOTICE %s :*** You cannot set %s Oper Only as it is currently "
					"Admin Only.", me.name, cptr->name, chptr->chname);
				break;
			}
		case MODE_NOCOLOUR:
		case MODE_STRIPCOLOUR:
			if (modetype == MODE_STRIPCOLOUR && (chptr->mode.mode & MODE_NOCOLOUR)) {
				chptr->mode.mode &= ~MODE_NOCOLOUR;
			}
			if (modetype == MODE_NOCOLOUR && (chptr->mode.mode & MODE_STRIPCOLOUR)) {
				chptr->mode.mode &= ~MODE_STRIPCOLOUR;
			}
		case MODE_PRIVATE:
		case MODE_SECRET:
		case MODE_MODERATED:
		case MODE_TOPICLIMIT:
		case MODE_NOPRIVMSGS:
		case MODE_INVITEONLY:
		case MODE_NONICKCH:
		case MODE_NOCTCP:
			retval = 0;
			if (what == MODE_ADD) {
				chptr->mode.mode |= modetype;
			}
			else {
				chptr->mode.mode &= ~modetype;
			}
			break;
		case MODE_CHANADMIN:
		case MODE_CHANOP:
		case MODE_HALFOP:
		case MODE_VOICE:
			if (param == NULL || *pcount >= MAXMODEPARAMS) {
				retval = 0;
				break;
			}

			retval = 1;
			if ((who = find_chasing(cptr, param, &chasing)) == NULL) {
				break;
			}
			if ((member = find_user_member(who->user->channel, chptr)) == NULL) {
				sendto_one(cptr, err_str(ERR_USERNOTINCHANNEL), me.name, cptr->name,
					who->name, chptr->chname);
				break;
			}
			switch (modetype) {
				case MODE_CHANADMIN:
					modename = "deadmin";
					modeflag = 'a';
					break;
				case MODE_CHANOP:
					modename = "deop";
					modeflag = 'o';
					break;
				case MODE_HALFOP:
					modename = "dehalfop";
					modeflag = 'h';
					break;
				case MODE_VOICE:
					modename = "devoice";
					modeflag = 'v';
					break;
			}
			if (modetype == MODE_CHANADMIN) {
				if (!IsServer(cptr) && !IsULine(cptr) && who != cptr && !HasMode(cptr, UMODE_NETADMIN)) {
					break;
				}
			}
			if (is_chan_admin(who, chptr) && what == MODE_DEL) {
				if (!IsULine(cptr) && !IsServer(cptr) && who != cptr && !HasMode(cptr, UMODE_NETADMIN)) {
					sendto_one(cptr, ":%s NOTICE %s :*** Cannot %s %s - user is channel admin.",
						me.name, cptr->name, modename, who->name);
					break;
				}
			}

			tmp = member->flags;
			if (what == MODE_ADD) {
				member->flags |= modetype;
			}
			else {
				member->flags &= ~modetype;
			}
			if (tmp == member->flags && (bounce == 1 || !IsULine(cptr))) {
				break;
			}
			if (bounce == 1) {
				member->flags = tmp;
			}
			ircsprintf(pvar[*pcount], "%c%c%s", what == MODE_ADD ? '+' : '-', modeflag, who->name);
			(*pcount)++;
			break;
		case MODE_LIMIT:
			if (what == MODE_ADD) {
				if (param == NULL) {
					retval = 0;
					break;
				}
				retval = 1;
				tmp = atoi(param);

				if (chptr->mode.limit == tmp) {
					break;
				}
				chptr->mode.limit = tmp;
			}
			else {
				retval = 0;
				if (chptr->mode.limit == 0) {
					break;
				}
				chptr->mode.limit = 0;
			}
			break;
		case MODE_KEY:
			if (param == NULL || *pcount >= MAXMODEPARAMS) {
				retval = 0;
				break;
			}
			retval = 1;

			for (x = 0; x < *pcount; x++) {
				if (pvar[x][1] == 'k') {
					retval = 0;
					break;
				}
			}
			if (retval == 0) {
				break;
			}
			if (what == MODE_ADD) {
				if (bounce == 0) {
					if (!check_chankey(param)) {
						break;
					}
					strncpyzt(chptr->mode.key, param, KEYLEN + 1);
				}
				tmpstr = param;
			}
			else {
				if (*chptr->mode.key == '\0') {
					break;
				}
				strncpyzt(tmpbuf, chptr->mode.key, sizeof(tmpbuf));
				tmpstr = tmpbuf;
				if (bounce == 1) {
					*chptr->mode.key = '\0';
					strcpy(chptr->mode.key, "");
				}
			}
			retval = 1;
			ircsprintf(pvar[*pcount], "%ck%s", what == MODE_ADD ? '+' : '-', tmpstr);
			(*pcount)++;
			break;
		case MODE_BAN:
			if (param == NULL || *pcount >= MAXMODEPARAMS) {
				retval = 0;
				break;
			}

			retval = 1;
			strcpy(nuhbuf, collapse(pretty_mask(param)));
			param = nuhbuf;
			if (bounce == 0) {
				if (what == MODE_ADD && !add_ban_tolist(cptr, chptr, param,
				  (aBan **)&chptr->banlist, 1)) {
					break;
				}
				else if (what == MODE_DEL && !del_ban_fromlist(chptr, param,
				  (aBan **)&chptr->banlist, 1)) {
					break;
				}
			}
			ircsprintf(pvar[*pcount], "%cb%s", what == MODE_ADD ? '+' : '-', nuhbuf);
			(*pcount)++;
			break;
		case MODE_EXBAN:
			if (param == NULL || *pcount >= MAXMODEPARAMS) {
				retval = 0;
				break;
			}

			retval = 1;
			strcpy(nuhbuf, collapse(pretty_mask(param)));
			param = nuhbuf;
			if (bounce == 0) {
				if (what == MODE_ADD && !add_ban_tolist(cptr, chptr, param,
				  (aBan **)&chptr->exbanlist, 0)) {
					break;
				}
				else if (what == MODE_DEL && !del_ban_fromlist(chptr, param,
				  (aBan **)&chptr->exbanlist, 0)) {
					break;
				}
			}
			ircsprintf(pvar[*pcount], "%ce%s", what == MODE_ADD ? '+' : '-', nuhbuf);
			(*pcount)++;
			break;
		case MODE_INVEX:
			if (param == NULL || *pcount >= MAXMODEPARAMS) {
				retval = 0;
				break;
			}

			retval = 1;
			strcpy(nuhbuf, collapse(pretty_mask(param)));
			param = nuhbuf;
			if (bounce == 0) {
				if (what == MODE_ADD && !add_ban_tolist(cptr, chptr, param,
				  (aBan **)&chptr->invexlist, 0)) {
					break;
				}
				else if (what == MODE_DEL && !del_ban_fromlist(chptr, param,
				  (aBan **)&chptr->invexlist, 0)) {
					break;
				}
			}
			ircsprintf(pvar[*pcount], "%cI%s", what == MODE_ADD ? '+' : '-', nuhbuf);
			(*pcount)++;
			break;
	}
	return retval;
}

void set_mode(aChannel *chptr, aClient *cptr, int parc, char *parv[], u_int *pcount,
	char pvar[MAXMODEPARAMS][MODEBUFLEN + 3], int bounce)
{
	int modetype = 0, found = 0, paracount = 1;
	char *curchr, *param = NULL;
	u_int what = MODE_ADD;
	aCtab *tab, foundat;

	*pcount = 0;
	for (curchr = parv[0]; *curchr; curchr++) {
		switch (*curchr) {
			case '+':
				what = MODE_ADD;
				break;
			case '-':
				what = MODE_DEL;
				break;
			default:
				found = 0;
				for (tab = &cFlagTab[0]; tab->mode != 0 && found == 0; tab++) {
					if (tab->flag == *curchr) {
						found = 1;
						foundat = *tab;
					}
				}
				if (found == 1) {
					modetype = foundat.mode;
				}
				if (found == 0) {
					modetype = 0;
				}
				if (modetype == 0) {
					if (MyClient(cptr)) {
						sendto_one(cptr, err_str(ERR_UNKNOWNMODE), me.name, cptr->name, *curchr);
					}
					break;
				}

				param = (parc - paracount >= 1) ? parv[paracount] : NULL;
				if (param != NULL && strlen(param) >= MODEBUFLEN) {
					param[MODEBUFLEN - 1] = '\0';
				}
				if (parv[paracount] != NULL && strlen(parv[paracount]) >= MODEBUFLEN) {
					parv[paracount][MODEBUFLEN - 1] = '\0';
				}
				paracount += do_mode_char(chptr, modetype, *curchr, param, what, cptr, pcount, pvar,
					bounce);
				break;
		}
	}
}

void make_mode_str(aChannel *chptr, u_int oldm, u_int oldl, int pcount, char pvar[MAXMODEPARAMS][MODEBUFLEN + 3],
	char *mode_buf, char *parabuf, char bounce)
{
	char *x = mode_buf, tmpbuf[MODEBUFLEN], *tmpstr;
	int what = 0, cnt, z;
	aCtab *tab;

	*tmpbuf = '\0';
	*mode_buf = '\0';
	*parabuf = '\0';

	for (tab = &cFlagTab[0]; tab->mode != 0; tab++) {
		if ((chptr->mode.mode & tab->mode) && !(oldm & tab->mode)) {
			if (what != MODE_ADD) {
				*x++ = bounce == 1 ? '-' : '+';
				what = MODE_ADD;
			}
			*x++ = tab->flag;
		}
	}
	*x = '\0';

	for (tab = &cFlagTab[0]; tab->mode != 0; tab++) {
		if (!(chptr->mode.mode & tab->mode) && (oldm & tab->mode)) {
			if (what != MODE_DEL) {
				*x++ = bounce == 1 ? '+' : '-';
				what = MODE_DEL;
			}
			*x++ = tab->flag;
		}
	}
	*x = '\0';

	for (cnt = 0; cnt < pcount; cnt++) {
		if ((*(pvar[cnt]) == '+') && what != MODE_ADD) {
			*x++ = bounce == 1 ? '-' : '+';
			what = MODE_ADD;
		}
		if ((*(pvar[cnt]) == '-') && what != MODE_DEL) {
			*x++ = bounce == 1 ? '+' : '-';
			what = MODE_DEL;
		}
		*x++ = *(pvar[cnt] + 1);
		tmpstr = &pvar[cnt][2];
		strncat(parabuf, tmpstr, MODEBUFLEN - 1);
		parabuf[MODEBUFLEN - 1] = '\0';
		z = strlen(parabuf);
		parabuf[z] = ' ';
		parabuf[z + 1] = '\0';
	}
	if (chptr->mode.limit != oldl) {
		if ((!bounce && chptr->mode.limit == 0) || (bounce && chptr->mode.limit != 0)) {
			if (what != MODE_DEL) {
				*x++ = '-';
				what = MODE_DEL;
			}
			if (bounce == 1) {
				chptr->mode.limit = 0;
			}
			*x++ = 'l';
		}
		else {
			if (what != MODE_ADD) {
				*x++ = '+';
				what = MODE_ADD;
			}
			if (bounce == 1) {
				chptr->mode.limit = oldl;
			}
			*x++ = 'l';
			ircsprintf(parabuf, "%s%d", parabuf, chptr->mode.limit);
		}
	}
	if (bounce == 1) {
		chptr->mode.mode = oldm;
	}

	z = strlen(parabuf);
	if (parabuf[z - 1] == ' ') {
		parabuf[z - 1] = '\0';
	}
	*x = '\0';

	if (*mode_buf == '\0') {
		*mode_buf = '+';
		mode_buf++;
		*mode_buf = '\0';
	}
	return;
}

void bounce_mode(aChannel *chptr, aClient *cptr, int parc, char *parv[])
{
	char mode_buf[MODEBUFLEN], parabuf[MODEBUFLEN];
	char pvar[MAXMODEPARAMS][MODEBUFLEN + 3];
	u_int oldm = chptr->mode.mode, oldl = chptr->mode.limit;
	int pcount;

	set_mode(chptr, cptr, parc, parv, &pcount, pvar, 1);
	make_mode_str(chptr, oldm, oldl, pcount, pvar, mode_buf, parabuf, 1);
	sendto_one(cptr, ":%s MODE %s &%s %s %lu", me.name, chptr->chname, mode_buf, parabuf, chptr->channelts);
	return;
}

void do_mode(aChannel *chptr, aClient *cptr, aClient *sptr, int parc, char *parv[], int sendts, int samode)
{
	char mode_buf[MODEBUFLEN], parabuf[MODEBUFLEN];
	char pvar[MAXMODEPARAMS][MODEBUFLEN + 3];
	char tschange = 0, isbounce = 0;
	u_int oldm = chptr->mode.mode, oldl = chptr->mode.limit;
	int pcount;

	if (**parv == '&') {
		isbounce = 1;
	}
	set_mode(chptr, sptr, parc, parv, &pcount, pvar, 0);
	make_mode_str(chptr, oldm, oldl, pcount, pvar, mode_buf, parabuf, 0);

	if (IsServer(sptr)) {
		if (sendts > 0) {
			if (sendts < chptr->channelts || chptr->channelts == 0) {
				tschange = 1;
				chptr->channelts = sendts;
			}
			if (sendts > chptr->channelts) {
				sendts = chptr->channelts;
				sendto_one(cptr, ":%s MODE %s + %lu", me.name, chptr->chname, chptr->channelts);
			}
		}
		if (sendts == -1 && chptr->channelts) {
			sendts = chptr->channelts;
		}
	}
	if (*mode_buf == '\0' || (*(mode_buf + 1) == '\0' && (*mode_buf == '+' || *mode_buf == '-'))) {
		if (tschange == 1 || isbounce == 1) {
			sendto_serv_butone(cptr, ":%s MODE %s %s+ %lu", me.name, chptr->chname,
				isbounce ? "&" : "", chptr->channelts);
		}
		return;
	}
	if (IsPerson(sptr) && samode == 1 && MyClient(sptr)) {
		sendto_serv_butone(NULL, ":%s GLOBOPS :%s used SAMODE %s (%s %s)", me.name, sptr->name,
			chptr->chname, mode_buf, parabuf);
		sendto_realops("%s used SAMODE %s (%s %s)", sptr->name, chptr->chname, mode_buf, parabuf);
		sendts = 0;
		used_samode = 1;
		sptr = &me;
	}
	sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s", sptr->name, chptr->chname, mode_buf, parabuf);

	if (IsServer(sptr) && sendts != -1) {
		sendto_serv_butone(cptr, ":%s MODE %s %s%s %s %lu", sptr->name, chptr->chname,
			isbounce ? "&" : "", mode_buf, parabuf);
	}
	else {
		sendto_serv_butone(cptr, ":%s MODE %s %s%s %s NoTS", sptr->name, chptr->chname,
			isbounce? "&" : "", mode_buf, parabuf);
	}
	return;
}

char *pretty_mask(char *mask)
{
	char *cp, *user, *host;

	if ((user = index((cp = mask), '!')) != NULL) {
		*user++ = '\0';
	}
	if ((host = rindex(user ? user : cp, '@'))) {
		*host++ = '\0';
		if (user == NULL) {
			return make_nick_user_host(NULL, cp, host);
		}
	}
	else if (user == NULL && index(cp, '.')) {
		return make_nick_user_host(NULL, NULL, cp);
	}
	return make_nick_user_host(cp, user, host);
}
