/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/ircd_signal.c: internal signal handling
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: ircd_signal.c,v 1.5 2002/10/26 20:54:15 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#ifdef HPUX
#define _KERNEL
#endif
#include <sys/resource.h>
#ifdef HPUX
#undef _KERNEL
#endif
#include <errno.h>
#include <pwd.h>
#include <signal.h>
#include <fcntl.h>
#include "inet.h"
#include "h.h"
#include "setup.h"

extern int do_rehash;

#ifdef PROFILING

#if defined(__GLIBC__) && (__GLIBC__ >= 2)
#include <sys/gmon.h>
#define monstartup __monstartup
#endif

extern void _start, etext;

static int profiling_state = 1;
static int profiling_newmsg = 0;
static char profiling_msg[512];

VOIDSIG s_dumpprof()
{
	char buf[32];
	ircsprintf(buf, "gmon.%d", (int)time(NULL));
	setenv("GMON_OUT_PREFIX", buf, 1);
	_mcleanup();
	monstartup((u_long)&_start, (u_long)&etext);
	setenv("GMON_OUT_PREFIX", "gmon.auto", 1);
	ircsprintf(profiling_msg, "Reset profile, saved past profile data to %s", buf);
	profiling_newmsg = 1;
}

VOIDSIG s_toggleprof()
{
	char buf[32];

	if (profiling_state == 1) {
		ircsprintf(buf, "gmon.%d", (int)time(NULL));
		setenv("GMON_OUT_PREFIX", buf, 1);
		_mcleanup();
		ircsprintf(profiling_msg, "Turned profiling OFF, saved profile data to %s", buf);
		profiling_state = 0;
	}
	else {
		monstartup((u_long)&_start, (u_long)&etext);
		setenv("GMON_OUT_PREFIX", "gmon.auto", 1);
		ircsprintf(profiling_msg, "Turned profiling ON");
		profiling_state = 1;
	}
	profiling_newmsg = 1;
}

#endif

VOIDSIG dummy()
{
#ifdef HAVE_RELIABLE_SIGNALS
	signal(SIGALRM, dummy);
	signal(SIGPIPE, dummy);
#ifdef SIGWINCH
	signal(SIGWINCH, dummy);
#endif
#else
#ifdef POSIX_SIGNALS
	struct sigaction act;
	act.sa_handler = dummy;
	act.sa_flags = 0;
	sigempty(&act.sa_mask);
	sigaddset(&act.sa_mask, SIGALRM);
	sigaddset(&act.sa_mask, SIGPIPE);
#ifdef SIGWINCH
	sigaddset(&act.sa_mask, SIGWINCH);
#endif
	sigaction(SIGALRM, &act, (struct sigaction *)NULL);
	sigaction(SIGPIPE, &act, (struct sigaction *)NULL);
#ifdef SIGWINCH
	sigaction(SIGWINCH, &act, (struct sigaction *)NULL);
#endif
#endif /* POSIX_SIGNALS */
#endif /* HAVE_RELIABLE_SIGNALS */
}

VOIDSIG s_die()
{
	dump_connections(me.fd);
	ircdlog(LOG_DEFAULT, "Server killed by SIGTERM");
	exit(0);
}

VOIDSIG s_rehash()
{
#ifdef POSIX_SIGNALS
	struct sigaction act;
#endif

	do_rehash = 1;

#ifdef POSIX_SIGNALS
	act.sa_handler = s_rehash;
	act.sa_flags = 0;
	sigemptyset(&act.sa_mask);
	sigaddset(&act.sa_mask, SIGHUP);
	sigaction(SIGHUP, &act, NULL);
#else
	signal(SIGHUP, s_rehash);
#endif
}

VOIDSIG s_restart()
{
	static int been_here = 0;
	if (been_here == 0) {
		ircdlog(LOG_DEFAULT, "Server restarting by SIGINT");
		been_here = 1;
		server_reboot();
	}
}

void init_signal()
{
#ifdef POSIX_SIGNALS
	struct sigaction act;

	act.sa_handler = SIG_IGN;
	act.sa_flags = 0;
	sigemptyset(&act.sa_mask);
	sigaddset(&act.sa_mask, SIGPIPE);
	sigaddset(&act.sa_mask, SIGALRM);
#ifdef SIGWINCH
	sigaddset(&act.sa_mask, SIGWINCH);
	sigaction(SIGWINCH, &act, NULL);
#endif
	sigaction(SIGPIPE, &act, NULL);
	act.sa_handler = dummy;
	sigaction(SIGALRM, &act, NULL);
	act.sa_handler = s_rehash;
	sigemptyset(&act.sa_mask);
	sigaddset(&act.sa_mask, SIGHUP);
	sigaction(SIGHUP, &act, NULL);
	act.sa_handler = s_restart;
	sigaddset(&act.sa_mask, SIGINT);
	sigaction(SIGINT, &act, NULL);
	act.sa_handler = s_die;
	sigaddset(&act.sa_mask, SIGTERM);
	sigaction(SIGTERM, &act, NULL);
#else /* POSIX_SIGNALS */
#ifndef HAVE_RELIABLE_SIGNALS
	signal(SIGPIPE, dummy);
#ifdef SIGWINCH
	signal(SIGWINCH, dummy);
#endif
#else /* HAVE_RELIABLE_SIGNALS */
#ifdef SIGWINCH
	signal(SIGWINCH, SIG_IGN);
#endif
	signal(SIGPIPE, SIG_IGN);
#endif /* HAVE_RELIABLE_SIGNALS */
	signal(SIGALRM, dummy);
	signal(SIGHUP, s_rehash);
	signal(SIGTERM, s_die);
	signal(SIGINT, s_restart);
#endif /* POSIX_SIGNALS */
#ifdef RESTARTING_SYSTEMCALL
	siginterrupt(SIGALRM, 1);
#endif
}

void init_signal_profiling()
{
#ifdef PROFILING
	setenv("GMON_OUT_PREFIX", "gmon.out", 1);
	signal(SIGUSR1, s_dumpprof);
	signal(SIGUSR2, s_toggleprof);
#endif
	return;
}

void check_profiling()
{
#ifdef PROFILING
	if (profiling_newmsg == 1) {
		sendto_realops("PROFILING: %s", profiling_msg);
		profiling_newmsg = 0;
	}
#endif
	return;
}
