/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/throttle.c: throttle system
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: throttle.c,v 1.13 2002/11/23 21:51:06 rageircd Exp $
 */

/*
 * Copyright 2000, 2001 Chip Norkus
 * Modified for RageIRCd v2 by Alasdair McWilliam. RageIRCd, Inc.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2a. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 * 2b. Redistribution in binary form requires specific prior written
 *     authorization of the maintainer.
 * 
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgement:
 *    This product includes software developed by Chip Norkus.
 * 
 * 4. The names of the maintainer, developers and contributors may not be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE MAINTAINER, DEVELOPERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE DEVELOPERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#include "numeric.h"
#include "blalloc.h"
#include "setup.h"
#include "queue.h"
#include "memory.h"
#include <sys/types.h>
#include <sys/socket.h>

#ifdef HAVE_THROTTLE_ON

int numthrottles = 0;
int numhashents = 0;
hash_table *throttle_hash;

SLIST_HEAD(hashent_list_t, hashent_t);
LIST_HEAD(throttle_list_t, throttle_t) throttles;

BlockHeap *throttle_heap;
BlockHeap *hashent_heap;

static char throttlebuf[512];
static int throttlelen = 0;

static hash_table *create_hash_table(int, size_t, size_t, int, int (*)(void *, void *));
static void resize_hash_table(hash_table *, int);
static unsigned int hash_get_key_hash(hash_table *, void *, size_t);
static int hash_insert(hash_table *, void *);
static int hash_delete(hash_table *, void *);
static void *hash_find(hash_table *, void *);

static hash_table *create_hash_table(int elems, size_t offset, size_t len, int flags, int (*cmpfunc)(void *, void *))
{
	hash_table *htp = (hash_table *)MyMalloc(sizeof(hash_table));
	htp->size = elems;
	htp->keyoffset = offset;
	htp->flags = flags;
	htp->cmpfunc = cmpfunc;
	htp->table = MyMallocX(sizeof(hashent_list) * htp->size);
	return htp;
}

static void resize_hash_table(hash_table *table, int elems)
{
	hashent_list *oldtable = table->table;
	hashent *hep;
	int oldsize = table->size, i;

	table->size = elems;
	table->table = MyMallocX(sizeof(hashent_list) * table->size);

	for (i = 0; i < oldsize; i++) {
		while (!SLIST_EMPTY(&oldtable[i])) {
			hep = SLIST_FIRST(&oldtable[i]);
			hash_insert(table, hep->ent);
			SLIST_REMOVE_HEAD(&oldtable[i], lp);
			free_hashent(hep);
		}
	}
	MyFree(oldtable);
	return;
}

static unsigned int hash_get_key_hash(hash_table *table, void *key, size_t offset)
{
	char *rkey = (char *)key + offset;
	int len = table->keylen;
	unsigned int hash = 0;

	if (len == 0) {
		len = strlen(rkey);
	}
	else if (table->flags & HASH_FL_STRING) {
		if ((len = strlen(rkey)) > table->keylen) {
			len = table->keylen;
		}
	}
	if (table->flags & HASH_FL_NOCASE) {
		while (len--) {
			hash = hash * 33 + tolower(*rkey++);
		}
	}
	else {
		while (len--) {
			hash = hash * 33 + *rkey++;
		}
	}
	return (hash % table->size);
}

static int hash_insert(hash_table *table, void *ent)
{
	int hash = hash_get_key_hash(table, ent, table->keyoffset);
	hashent *hep = make_hashent();
	hep->ent = ent;
	SLIST_INSERT_HEAD(&table->table[hash], hep, lp);
	numhashents++;
	return 1;
}

static int hash_delete(hash_table *table, void *ent)
{
	int hash = hash_get_key_hash(table, ent, table->keyoffset);
	hashent *hep;

	SLIST_FOREACH(hep, &table->table[hash], lp) {
		if (hep->ent == ent) {
			break;
		}
	}
	if (hep == NULL) {
		return 0;
	}
	SLIST_REMOVE(&table->table[hash], hep, hashent_t, lp);
	free_hashent(hep);
	numhashents--;
	return 1;
}

static void *hash_find(hash_table *table, void *key)
{
	int hash = hash_get_key_hash(table, key, 0);
	hashent *hep;

	SLIST_FOREACH(hep, &table->table[hash], lp) {
		if (!table->cmpfunc(&((char *)hep->ent)[table->keyoffset], key)) {
			return hep->ent;
		}
	}
	return NULL;
}

void init_throttle(void)
{
	throttle_heap = BlockHeapCreate((size_t)sizeof(throttle), THROTTLES_PREALLOCATE);
	hashent_heap = BlockHeapCreate((size_t)sizeof(hashent), HASHENTS_PREALLOCATE);
	throttle_hash = create_hash_table(THROTTLE_HASHSIZE, offsetof(throttle, addr), HOSTIPLEN,
		HASH_FL_STRING, (int (*)(void *, void *))irccmp);
	throttlelen = ircsnprintf(throttlebuf, 512, "ERROR :Your host is trying to (re)connect too fast -- throttled.\r\n");
	return;
}

static int throttle_get_zline_time(int stage)
{
	switch (stage) {
		case -1:
			return 0;
		case 0:
			return 120;
		case 1:
			return 300;
		case 2:
			return 900;
		case 3:
			return 1800;
		default:
			return 3600;
	}
	return 0;
}

void throttle_remove(char *host)
{
	throttle *tp;
	if ((tp = hash_find(throttle_hash, host)) != NULL) {
		LIST_REMOVE(tp, lp);
		hash_delete(throttle_hash, tp);
		free_throttle(tp);
		numthrottles--;
	}
	return;
}

int throttle_force(char *host)
{
	throttle *tp;

	if (THROTTLE_COUNT < 1 || THROTTLE_TIME < 1) {
		return 0;
	}
	if ((tp = hash_find(throttle_hash, host)) == NULL) {
		tp = make_throttle();
		strcpy(tp->addr, host);
		tp->stage = -1;
		tp->zline_start = 0;
		tp->conns = 0;
		tp->first = timeofday;
		tp->re_zlines = 0;

		hash_insert(throttle_hash, tp);
		LIST_INSERT_HEAD(&throttles, tp, lp);
		numthrottles++;
	}
	tp->conns = THROTTLE_COUNT;
	tp->last = tp->first = timeofday;
	return 1;
}

int throttle_check(char *host, int fd, time_t sotime)
{
	throttle *tp;

	if (THROTTLE_COUNT < 1 || THROTTLE_TIME < 1) {
		return 1;
	}
	if (fd == -1 && (timeofday - sotime > THROTTLE_TIME)) {
		return 1;
	}
	if (sotime > timeofday) {
		sotime = timeofday;
	}
	if ((tp = hash_find(throttle_hash, host)) == NULL) {
		tp = make_throttle();
		strcpy(tp->addr, host);
		tp->stage = -1;
		tp->zline_start = 0;
		tp->conns = 0;
		tp->first = sotime;
		tp->re_zlines = 0;

		hash_insert(throttle_hash, tp);
		LIST_INSERT_HEAD(&throttles, tp, lp);
		numthrottles++;
	}
	else if (tp->zline_start != 0) {
		time_t zlength = throttle_get_zline_time(tp->stage);

		if (sotime - tp->zline_start < zlength) {
			if (fd == -1) {
				return 0;
			}
			send(fd, throttlebuf, throttlelen, 0);
			tp->re_zlines++;
			tp->zline_start = sotime;
			return 0;
		}
		tp->conns = 0;
		tp->first = sotime;
		tp->zline_start = 0;
	}

	tp->conns++;
	tp->last = sotime;

	if (sotime - tp->first > THROTTLE_TIME) {
		tp->conns = 1;
		tp->first = sotime;
		return 1;
	}
	if (tp->conns >= THROTTLE_COUNT) {
		if (fd != -1) {
			char errbuf[512];
			int elength, zlength = throttle_get_zline_time(++tp->stage);

			sendto_realops_lev(CCONN_LEV, "throttled connections from %s (%d in %d seconds) for %d "
				"minutes (offense %d)", tp->addr, tp->conns, sotime - tp->first, zlength / 60,
				tp->stage + 1);
			elength = ircsnprintf(errbuf, 512, ":%s NOTICE ZUSR :You have been throttled for %d "
				"minutes for too many connections in a short period of time. Further "
				"connections during this period will reset your throttle and you will "
				"have to wait longer.\r\n", me.name, zlength / 60);
			send(fd, errbuf, elength, 0);

			if (throttle_get_zline_time(tp->stage + 1) != zlength) {
				elength = ircsnprintf(errbuf, 512, ":%s NOTICE ZUSR :When you reconnect, if "
					"you are throttled again, it will last longer.\r\n", me.name);
				send(fd, errbuf, elength, 0);
			}

			send(fd, throttlebuf, throttlelen, 0);
			tp->zline_start = sotime;
		}
		return 0;
	}
	return 1;
}

void throttle_timer(time_t now)
{
	throttle *tp, *tp2;
	time_t zlength;

	if (THROTTLE_COUNT < 1 || THROTTLE_TIME < 1) {
		return;
	}
	for (tp = LIST_FIRST(&throttles); tp; tp = tp2) {
		tp2 = LIST_NEXT(tp, lp);
		zlength = throttle_get_zline_time(tp->stage);
		if (now == 0 || (tp->zline_start && (now - tp->zline_start) >= (zlength + THROTTLE_RECORDTIME)) ||
		  (tp->zline_start == 0 && (now - tp->last) >= THROTTLE_RECORDTIME)) {
			LIST_REMOVE(tp, lp);
			hash_delete(throttle_hash, tp);
			free_throttle(tp);
			numthrottles--;
		}
	}
	return;
}

void throttle_rehash(void)
{
	throttle_timer(0);
	return;
}
void throttle_resize(int size)
{
	resize_hash_table(throttle_hash, size);
	return;
}

void throttle_stats(aClient *cptr)
{
	int pending = 0, bans = 0;
	throttle *tp;

	sendto_one(cptr, ":%s %d %s :throttles: %d", me.name, RPL_STATSDEBUG, cptr->name, numthrottles);
	sendto_one(cptr, ":%s %d %s :memory allocation: %d hashents (%d bytes) %d throttles (%d bytes)",
		me.name, RPL_STATSDEBUG, cptr->name, numhashents, numhashents * sizeof(hashent),
		numthrottles, numthrottles * sizeof(throttle));
	sendto_one(cptr, ":%s %d %s :throttle hash table size: %d", me.name, RPL_STATSDEBUG, cptr->name,
		throttle_hash->size);

	LIST_FOREACH(tp, &throttles, lp) {
		if (tp->zline_start > 0) {
			bans++;
		}
		else {
			pending++;
		}
	}
	sendto_one(cptr, ":%s %d %s :throttles pending=%d bans=%d", me.name, RPL_STATSDEBUG, cptr->name,
		pending, bans);

	LIST_FOREACH(tp, &throttles, lp) {
		int ztime = throttle_get_zline_time(tp->stage);
		if (tp->zline_start && tp->zline_start + ztime > timeofday) {
			sendto_one(cptr, ":%s %d %s :throttled: %s [stage %d, %d secs remain, %d retries]",
				me.name, RPL_STATSDEBUG, cptr->name, tp->addr, tp->stage,
				(tp->zline_start + ztime) - timeofday, tp->re_zlines);
		}
	}
	return;
}

#endif
