/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/whowas.c: whowas system
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: whowas.c,v 1.8 2002/11/16 21:12:28 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "h.h"
#include "memory.h"

/* externally defined functions */
unsigned int hash_whowas_name(char *);

/* internally defined function */
static void add_whowas_to_clist(aWhowas **, aWhowas *);
static void del_whowas_from_clist(aWhowas **, aWhowas *);
static void add_whowas_to_list(aWhowas **, aWhowas *);
static void del_whowas_from_list(aWhowas **, aWhowas *);

aWhowas WHOWAS[NICKNAMEHISTORYLENGTH];
aWhowas *WHOWASHASH[WW_MAX];
int whowas_next = 0;

void add_history(aClient *cptr, int online)
{
	aWhowas *new;

	for (new = WHOWASHASH[hash_whowas_name(cptr->name)]; new; new = new->next) {
		if (!mycmp(cptr->name, new->name) && !mycmp(cptr->username, new->username) &&
		  !mycmp(cptr->username, new->username) && !mycmp(cptr->host, new->hostname)
		  && !mycmp(cptr->user->maskedhost, new->maskedhost) && !mycmp(cptr->info, new->realname)) {
			new->logoff = timeofday;
			return;
		}
	}

	new = &WHOWAS[whowas_next];
	if (new->hashv != -1) {
		if (new->online != NULL) {
			del_whowas_from_clist(&(new->online->user->whowas), new);
		}
		del_whowas_from_list(&WHOWASHASH[new->hashv], new);
	}

	new->hashv = hash_whowas_name(cptr->name);
	new->logoff = timeofday;
	strncpyzt(new->name, cptr->name, NICKLEN + 1);
	strncpyzt(new->username, cptr->username, USERLEN + 1);
	strncpyzt(new->hostname, cptr->host, HOSTLEN + 1);
	strncpyzt(new->maskedhost, MaskedHost(cptr), HOSTLEN + 1);
	strncpyzt(new->realname, cptr->info, REALLEN + 1);
	new->servername = cptr->user->server;

	if (online == 1) {
		new->online = cptr;
		add_whowas_to_clist(&(cptr->user->whowas), new);
	}
	else {
		new->online = NULL;
	}
	add_whowas_to_list(&WHOWASHASH[new->hashv], new);
	whowas_next++;
	if (whowas_next == NICKNAMEHISTORYLENGTH) {
		whowas_next = 0;
	}
	return;
}

void off_history(aClient *cptr)
{
	aWhowas *tmp, *next;

	for (tmp = cptr->user->whowas; tmp; tmp = next) {
		next = tmp->cnext;
		tmp->online = NULL;
		del_whowas_from_clist(&(cptr->user->whowas), tmp);
	}
}

aClient *get_history(char *nick, time_t timelimit)
{
	int blah = hash_whowas_name(nick);
	aWhowas *tmp = WHOWASHASH[blah];

	timelimit = timeofday - timelimit;
	for (; tmp; tmp = tmp->next) {
		if (mycmp(nick, tmp->name) || tmp->logoff < timelimit) {
			continue;
		}
		return tmp->online;
	}
	return NULL;
}

void count_whowas_memory(int *wwu, u_long *wwum)
{
	aWhowas *tmp;
	int i, u = 0;
	u_long um = 0;

	for (i = 0, tmp = &WHOWAS[0]; i < NICKNAMEHISTORYLENGTH; i++, tmp++) {
		if (tmp->hashv != -1) {
			u++;
			um += sizeof(aWhowas);
		}
	}
	*wwu = u;
	*wwum = um;
	return;
}

void init_whowas()
{
	int i;
	for (i = 0; i < NICKNAMEHISTORYLENGTH; i++) {
		memset((char *)&WHOWAS[i], '\0', sizeof(aWhowas));
		WHOWAS[i].hashv = -1;
	}
	for (i = 0; i < WW_MAX; i++) {
		WHOWASHASH[i] = NULL;
	}
}

static void add_whowas_to_clist(aWhowas **bucket, aWhowas *whowas)
{
	whowas->cprev = NULL;
	if ((whowas->cnext = *bucket) != NULL) {
		whowas->cnext->cprev = whowas;
	}
	*bucket = whowas;
}

static void del_whowas_from_clist(aWhowas **bucket, aWhowas *whowas)
{
	if (whowas->cprev) {
		whowas->cprev->cnext = whowas->cnext;
	}
	else {
		*bucket = whowas->cnext;
	}
	if (whowas->cnext) {
		whowas->cnext->cprev = whowas->cprev;
	}
}

static void add_whowas_to_list(aWhowas **bucket, aWhowas *whowas)
{
	whowas->prev = NULL;
	if ((whowas->next = *bucket) != NULL) {
		whowas->next->prev = whowas;
	}
	*bucket = whowas;
}

static void del_whowas_from_list(aWhowas **bucket, aWhowas *whowas)
{
	if (whowas->prev) {
		whowas->prev->next = whowas->next;
	}
	else {
		*bucket = whowas->next;
	}
	if (whowas->next) {
		whowas->next->prev = whowas->prev;
	}
}
